#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Automatically fix lintian issues."""

from datetime import datetime
import errno
import io
import itertools
import logging
import os
import re
import subprocess
import sys
import tempfile
import time
import traceback
from typing import Optional, List, Sequence, Iterator, Iterable, Tuple

from tqdm import trange

from debian.changelog import Changelog, Version

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401
from breezy.commit import NullCommitReporter
from breezy.errors import NoSuchFile
from breezy.osutils import is_inside
from breezy.rename_map import RenameMap
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.workspace import reset_tree, check_clean_tree

from debian.deb822 import Deb822


from debmutate.reformatting import FormattingUnpreservable


__version__ = (0, 120)
version_string = ".".join(map(str, __version__))
SUPPORTED_CERTAINTIES = ["certain", "confident", "likely", "possible", None]
DEFAULT_MINIMUM_CERTAINTY = "certain"
USER_AGENT = "lintian-brush/" + version_string
# Too aggressive?
DEFAULT_URLLIB_TIMEOUT = 3
logger = logging.getLogger(__name__)


class NoChanges(Exception):
    """Script didn't make any changes."""

    def __init__(self, fixer, comment=None, overridden_lintian_issues=None):
        super(NoChanges, self).__init__(fixer, comment)
        self.fixer = fixer
        self.overridden_lintian_issues = overridden_lintian_issues or []


class NotCertainEnough(NoChanges):
    """Script made changes but with too low certainty."""

    def __init__(self, fixer, certainty, minimum_certainty,
                 overridden_lintian_issues=None):
        super(NotCertainEnough, self).__init__(
            fixer, overridden_lintian_issues=overridden_lintian_issues)
        self.certainty = certainty
        self.minimum_certainty = minimum_certainty


class FixerFailed(Exception):
    """Base class for fixer script failures."""

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.args == other.args


class UnsupportedCertainty(Exception):
    """Unsupported certainty."""


class FixerScriptFailed(FixerFailed):
    """Script failed to run."""

    def __init__(self, path, returncode, errors):
        self.path = path
        self.returncode = returncode
        self.errors = errors

    def __str__(self):
        return "Script %s failed with exit code: %d\n%s\n" % (
            self.path,
            self.returncode,
            self.errors,
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.path == other.path
            and self.returncode == other.returncode
            and self.errors == other.errors
        )


class DescriptionMissing(Exception):
    """The fixer script did not provide a description on stdout."""

    def __init__(self, fixer):
        super(DescriptionMissing, self).__init__(fixer)
        self.fixer = fixer


class NotDebianPackage(Exception):
    """The specified directory does not contain a Debian package."""

    def __init__(self, tree, path):
        super(NotDebianPackage, self).__init__(tree.abspath(path))


class FixerResult(object):
    """Result of a fixer run."""

    def __init__(
        self,
        description,
        fixed_lintian_tags=[],
        certainty=None,
        patch_name=None,
        revision_id=None,
        fixed_lintian_issues=[],
        overridden_lintian_issues=[],
    ):
        self.description = description
        self.fixed_lintian_issues = fixed_lintian_issues or []
        if fixed_lintian_tags:
            self.fixed_lintian_issues.extend(
                [LintianIssue(tag=tag) for tag in fixed_lintian_tags])
        self.overridden_lintian_issues = overridden_lintian_issues
        self.certainty = certainty
        self.patch_name = patch_name
        self.revision_id = revision_id

    @property
    def fixed_lintian_tags(self):
        return [issue.tag for issue in self.fixed_lintian_issues]

    def __repr__(self):
        return (
            "%s(%r, fixed_lintian_issues=%r, "
            "overridden_lintian_issues=%r, certainty=%r, patch_name=%r, "
            "revision_id=%r)"
        ) % (
            self.__class__.__name__,
            self.description,
            self.fixed_lintian_issues,
            self.overridden_lintian_issues,
            self.certainty,
            self.patch_name,
            self.revision_id,
        )

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return (
            (self.description == other.description)
            and (self.fixed_lintian_issues == other.fixed_lintian_issues)
            and (self.overridden_lintian_issues == other.overridden_lintian_issues)
            and (self.certainty == other.certainty)
            and (self.patch_name == other.patch_name)
            and (self.revision_id == other.revision_id)
        )


class Fixer(object):
    """A Fixer script.

    The `lintian_tags` attribute contains the name of the lintian tags this
    fixer addresses.
    """

    def __init__(self, name: str, lintian_tags: List[str] = None):
        self.name = name
        self.lintian_tags = lintian_tags or []

    def run(
        self,
        basedir,
        package,
        current_version,
        compat_release,
        minimum_certainty=None,
        trust_package=False,
        allow_reformatting=False,
        net_access=True,
        opinionated=False,
        diligence=0,
    ):
        """Apply this fixer script.

        Args:
          basedir: Directory in which to run
          package: Name of the source package
          current_version: The version of the package that is being created or
            updated
          compat_release: Compatibility level (a Debian release name)
          trust_package: Whether to run code from the package
          allow_reformatting: Allow reformatting of files that are being
            changed
          opinionated: Whether to be opinionated
          diligence: Level of diligence
        Returns:
          A FixerResult object
        """
        raise NotImplementedError(self.run)


class LintianIssue(object):

    def __init__(
            self,
            tag: str,
            package: Optional[str] = None,
            package_type: Optional[str] = None,
            info: Optional[List[str]] = None):
        self.package = package
        self.package_type = package_type
        self.tag = tag
        self.info = info

    def __eq__(self, other):
        return isinstance(self, type(other)) and (
            self.package == other.package and
            self.package_type == other.package_type and
            self.tag == other.tag and
            self.info == other.info)

    @classmethod
    def from_str(cls, text):
        try:
            (before, after) = text.strip().split(':', 1)
        except ValueError:
            package_type = package = None
            after = text
        else:
            try:
                (package_type, package) = before.strip().split(' ')
            except ValueError:
                package = before
                package_type = None
        parts = after.strip().split(' ')
        return cls(
            package=package,
            package_type=package_type,
            tag=parts[0],
            info=parts[1:])


def parse_script_fixer_output(text):
    """Parse the output from a script fixer."""
    description = []
    overridden_issues = []
    fixed_issues = []
    fixed_tags = []
    certainty = None
    patch_name = None
    lines = text.splitlines()
    i = 0
    while i < len(lines):
        # TODO(jelmer): Do this in a slightly less hackish manner
        try:
            (key, value) = lines[i].split(":", 1)
        except ValueError:
            description.append(lines[i])
        else:
            if key == "Fixed-Lintian-Tags":
                fixed_tags.extend(
                    [tag.strip() for tag in value.strip().split(",")])
            elif key == "Fixed-Lintian-Issues":
                i += 1
                while i < len(lines) and lines[i].startswith(' '):
                    fixed_issues.append(LintianIssue.from_str(lines[i][1:]))
                    i += 1
                continue
            elif key == "Overridden-Lintian-Issues":
                i += 1
                while i < len(lines) and lines[i].startswith(' '):
                    overridden_issues.append(
                        LintianIssue.from_str(lines[i][1:]))
                    i += 1
                continue
            elif key == "Certainty":
                certainty = value.strip()
            elif key == "Patch-Name":
                patch_name = value.strip()
            else:
                description.append(lines[i])
        i += 1
    if certainty not in SUPPORTED_CERTAINTIES:
        raise UnsupportedCertainty(certainty)
    return FixerResult(
        "\n".join(description), fixed_tags,
        certainty, patch_name, revision_id=None,
        fixed_lintian_issues=fixed_issues,
        overridden_lintian_issues=overridden_issues)


def determine_env(
    package,
    current_version,
    compat_release,
    minimum_certainty,
    trust_package,
    allow_reformatting,
    net_access,
    opinionated,
    diligence,
):
    env = dict(os.environ.items())
    env["DEB_SOURCE"] = package
    env["CURRENT_VERSION"] = str(current_version)
    env["COMPAT_RELEASE"] = compat_release
    env["MINIMUM_CERTAINTY"] = minimum_certainty
    env["TRUST_PACKAGE"] = "true" if trust_package else "false"
    env["REFORMATTING"] = "allow" if allow_reformatting else "disallow"
    env["NET_ACCESS"] = "allow" if net_access else "disallow"
    env["OPINIONATED"] = "yes" if opinionated else "no"
    env["DILIGENCE"] = str(diligence)
    return env


class PythonScriptFixer(Fixer):
    """A fixer that is implemented as a python script.

    This gets used just for Python scripts, and significantly speeds
    things up because it prevents starting a new Python interpreter
    for every fixer.
    """

    def __init__(self, name, lintian_tags, script_path):
        super(PythonScriptFixer, self).__init__(name, lintian_tags)
        self.script_path = script_path

    def __repr__(self):
        return "<%s(%r)>" % (self.__class__.__name__, self.name)

    def __str__(self):
        return self.name

    def run(
        self,
        basedir,
        package,
        current_version,
        compat_release,
        minimum_certainty=DEFAULT_MINIMUM_CERTAINTY,
        trust_package=False,
        allow_reformatting=False,
        net_access=True,
        opinionated=False,
        diligence=0,
    ):
        env = determine_env(
            package=package,
            current_version=current_version,
            compat_release=compat_release,
            minimum_certainty=minimum_certainty,
            trust_package=trust_package,
            allow_reformatting=allow_reformatting,
            net_access=net_access,
            opinionated=opinionated,
            diligence=diligence,
        )
        try:
            old_env = os.environ
            old_stderr = sys.stderr
            old_stdout = sys.stdout
            sys.stderr = io.StringIO()
            sys.stdout = io.StringIO()
            os.environ = env
            try:
                old_cwd = os.getcwd()
            except FileNotFoundError:
                old_cwd = None
            try:
                os.chdir(basedir)
                global_vars = {
                    "__file__": self.script_path,
                    "__name__": "__main__",
                }
                with open(self.script_path, "r") as f:
                    code = compile(f.read(), self.script_path, "exec")
                    exec(code, global_vars)
            except FormattingUnpreservable:
                raise
            except SystemExit as e:
                retcode = e.code
            except BaseException as e:
                traceback.print_exception(type(e), e, e.__traceback__, file=sys.stderr)
                raise FixerScriptFailed(self.script_path, 1, sys.stderr.getvalue())
            else:
                retcode = 0
            description = sys.stdout.getvalue()
            err = sys.stderr.getvalue()
        finally:
            os.environ = old_env
            sys.stderr = old_stderr
            sys.stdout = old_stdout
            if old_cwd is not None:
                os.chdir(old_cwd)
            from . import fixer

            fixer.reset()

        if retcode == 2:
            raise NoChanges(self)
        if retcode != 0:
            raise FixerScriptFailed(self.script_path, retcode, err)

        return parse_script_fixer_output(description)


class ScriptFixer(Fixer):
    """A fixer that is implemented as a shell/python/etc script."""

    def __init__(self, name: str, lintian_tags: List[str], script_path: str):
        super(ScriptFixer, self).__init__(name, lintian_tags)
        self.script_path = script_path

    def __repr__(self):
        return "<ScriptFixer(%r)>" % self.name

    def __str__(self):
        return self.name

    def run(
        self,
        basedir: str,
        package: str,
        current_version: Version,
        compat_release: str,
        minimum_certainty: str = DEFAULT_MINIMUM_CERTAINTY,
        trust_package: bool = False,
        allow_reformatting: bool = False,
        net_access: bool = True,
        opinionated: bool = False,
        diligence: int = 0,
    ):
        env = determine_env(
            package=package,
            current_version=current_version,
            compat_release=compat_release,
            minimum_certainty=minimum_certainty,
            trust_package=trust_package,
            allow_reformatting=allow_reformatting,
            net_access=net_access,
            opinionated=opinionated,
            diligence=diligence,
        )
        with tempfile.SpooledTemporaryFile() as stderr:
            try:
                p = subprocess.Popen(
                    self.script_path,
                    cwd=basedir,
                    stdout=subprocess.PIPE,
                    stderr=stderr,
                    env=env,
                )
            except OSError as e:
                if e.errno == errno.ENOMEM:
                    raise MemoryError
                raise
            (description, err) = p.communicate(b"")
            if p.returncode == 2:
                raise NoChanges(self)
            if p.returncode != 0:
                stderr.seek(0)
                raise FixerScriptFailed(
                    self.script_path,
                    p.returncode,
                    stderr.read().decode("utf-8", "replace"),
                )
        return parse_script_fixer_output(description.decode("utf-8"))


def find_fixers_dir() -> str:
    """Find the local directory with lintian fixer scripts."""
    local_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "fixers"))
    if os.path.isdir(local_dir):
        return local_dir
    import pkg_resources

    resource_dir = pkg_resources.resource_filename(__name__, "lintian-brush/fixers")
    if os.path.isdir(resource_dir):
        return resource_dir
    # Urgh.
    local_share_dir = "/usr/local/share/lintian-brush/fixers"
    if os.path.isdir(local_share_dir):
        return local_share_dir

    return "/usr/share/lintian-brush/fixers"


def read_desc_file(path: str, force_subprocess: bool = False) -> Iterator[Fixer]:
    """Read a description file.

    Args:
      path: Path to read from.
      force_subprocess: Force running as subprocess
    Yields:
      Fixer objects
    """
    dirname = os.path.dirname(path)
    with open(path, "r") as f:
        for paragraph in Deb822.iter_paragraphs(f):
            name = os.path.splitext(paragraph["Fix-Script"])[0]
            script_path = os.path.join(dirname, paragraph["Fix-Script"])
            if "Lintian-Tags" in paragraph:
                tags = [tag.strip() for tag in paragraph["Lintian-Tags"].split(",")]
            else:
                tags = []
            if script_path.endswith(".py") and not force_subprocess:
                yield PythonScriptFixer(name, tags, script_path)
            else:
                yield ScriptFixer(name, tags, script_path)


def select_fixers(
    fixers: List[Fixer], names: List[str], exclude: Optional[Iterable[str]] = None
) -> List[Fixer]:
    """Select fixers by name, from a list.

    Args:
      fixers: List of Fixer objects
      names: Set of names to select
      exclude: Set of names to exclude
    Raises:
      KeyError: if one of the names did not exist
    """
    names_set = set(names)
    if exclude:
        for name in exclude:
            if name not in names_set:
                raise KeyError(name)
            names_set.remove(name)
    available = set([f.name for f in fixers])
    missing = names_set - available
    if missing:
        raise KeyError(missing.pop())
    # Preserve order
    return [f for f in fixers if f.name in names_set]


def available_lintian_fixers(
    fixers_dir: Optional[str] = None, force_subprocess: bool = False
) -> Iterator[Fixer]:
    """Return a list of available lintian fixers.

    Args:
      fixers_dir: Fixers directory to browse
      force_subprocess: Force running fixers from subprocess
    Returns:
      Iterator over Fixer objects
    """
    if fixers_dir is None:
        fixers_dir = find_fixers_dir()
    for n in os.listdir(fixers_dir):
        if not n.endswith(".desc"):
            continue
        for fixer in read_desc_file(
            os.path.join(fixers_dir, n), force_subprocess=force_subprocess
        ):
            yield fixer


def increment_version(v: Version) -> None:
    """Increment a version number.

    For native packages, increment the main version number.
    For other packages, increment the debian revision.

    Args:
        v: Version to increment (modified in place)
    """
    if v.debian_revision is not None:
        v.debian_revision = re.sub(
            "\\d+$", lambda x: str(int(x.group()) + 1), v.debian_revision
        )
    else:
        v.upstream_version = re.sub(
            "\\d+$", lambda x: str(int(x.group()) + 1), v.upstream_version
        )


def get_committer(tree: WorkingTree) -> str:
    """Get the committer string for a tree.

    Args:
      tree: A Tree object
    Returns:
      A committer string
    """
    # TODO(jelmer): Perhaps this logic should be in Breezy?
    if getattr(tree.branch.repository, "_git", None):
        cs = tree.branch.repository._git.get_config_stack()
        user = os.environ.get("GIT_COMMITTER_NAME")
        email = os.environ.get("GIT_COMMITTER_EMAIL")
        if user is None:
            try:
                user = cs.get(("user",), "name").decode("utf-8")
            except KeyError:
                user = None
        if email is None:
            try:
                email = cs.get(("user",), "email").decode("utf-8")
            except KeyError:
                email = None
        if user and email:
            return user + " <" + email + ">"
        from breezy.config import GlobalStack

        return GlobalStack().get("email")
    else:
        config = tree.branch.get_config_stack()
        return config.get("email")


def only_changes_last_changelog_block(
    tree: WorkingTree, basis_tree: Tree, changelog_path: str, changes
) -> bool:
    """Check whether the only change in a tree is to the last changelog entry.

    Args:
      tree: Tree to analyze
      changelog_path: Path to the changelog file
      changes: Changes in the tree
    Returns:
      boolean
    """
    with tree.lock_read(), basis_tree.lock_read():
        changes_seen = False
        for change in changes:
            if change.path[1] == "":
                continue
            if change.path[1] == changelog_path:
                changes_seen = True
                continue
            if not tree.has_versioned_directories() and is_inside(
                change.path[1], changelog_path
            ):
                continue
            return False
        if not changes_seen:
            return False
        try:
            new_cl = Changelog(tree.get_file_text(changelog_path))
        except NoSuchFile:
            return False
        try:
            old_cl = Changelog(basis_tree.get_file_text(changelog_path))
        except NoSuchFile:
            return True
        if old_cl.distributions != "UNRELEASED":
            return False
        del new_cl._blocks[0]
        del old_cl._blocks[0]
        return str(new_cl) == str(old_cl)


def certainty_sufficient(
    actual_certainty: str, minimum_certainty: Optional[str]
) -> bool:
    """Check if the actual certainty is sufficient.

    Args:
      actual_certainty: Actual certainty with which changes were made
      minimum_certainty: Minimum certainty to keep changes
    Returns:
      boolean
    """
    actual_confidence = certainty_to_confidence(actual_certainty)
    if actual_confidence is None:
        # Actual confidence is unknown.
        # TODO(jelmer): Should we really be ignoring this?
        return True
    minimum_confidence = certainty_to_confidence(minimum_certainty)
    if minimum_confidence is None:
        return True
    return actual_confidence <= minimum_confidence


def has_non_debian_changes(changes, subpath):
    for change in changes:
        for path in change.path:
            if path and not is_inside(os.path.join(subpath, "debian"), path):
                return True
    return False


_changelog_policy_noted = False


def _note_changelog_policy(policy, msg):
    global _changelog_policy_noted
    if not _changelog_policy_noted:
        if policy:
            extra = "Specify --no-update-changelog to override."
        else:
            extra = "Specify --update-changelog to override."
        logging.info("%s %s", msg, extra)
    _changelog_policy_noted = True


class FailedPatchManipulation(Exception):
    def __init__(self, tree, patches_directory, reason):
        super(FailedPatchManipulation, self).__init__(tree, patches_directory, reason)


def _upstream_changes_to_patch(
    local_tree: WorkingTree,
    basis_tree: Tree,
    dirty_tracker,
    subpath: str,
    patch_name: str,
    patch_description: str,
    timestamp: Optional[datetime] = None,
) -> Tuple[str, List[str]]:
    from .patches import (
        move_upstream_changes_to_patch,
        read_quilt_patches,
        tree_patches_directory,
        PatchSyntax,
    )

    # TODO(jelmer): Apply all patches before generating a diff.

    patches_directory = tree_patches_directory(local_tree, subpath)
    try:
        quilt_patches = list(read_quilt_patches(local_tree, patches_directory))
    except PatchSyntax as e:
        raise FailedPatchManipulation(
            local_tree, patches_directory, "Unable to parse some patches: %s" % e
        )
    if len(quilt_patches) > 0:
        raise FailedPatchManipulation(
            local_tree,
            patches_directory,
            "Creating patch on top of existing upstream " "patches not supported.",
        )

    logging.debug("Moving upstream changes to patch %s", patch_name)
    try:
        specific_files, patch_name = move_upstream_changes_to_patch(
            local_tree,
            basis_tree,
            subpath,
            patch_name,
            patch_description,
            dirty_tracker,
            timestamp=timestamp,
        )
    except FileExistsError as e:
        raise FailedPatchManipulation(
            local_tree, patches_directory, "patch path %s already exists\n" % e.args[0]
        )

    return patch_name, specific_files


def run_lintian_fixer(  # noqa: C901
    local_tree: WorkingTree,
    fixer: Fixer,
    committer: Optional[str] = None,
    update_changelog: Optional[bool] = None,
    compat_release: Optional[str] = None,
    minimum_certainty: Optional[str] = None,
    trust_package: bool = False,
    allow_reformatting: bool = False,
    dirty_tracker=None,
    subpath: str = "",
    net_access: bool = True,
    opinionated: Optional[bool] = None,
    diligence: int = 0,
    timestamp: Optional[datetime] = None,
    basis_tree: Optional[Tree] = None,
    changes_by: str = "lintian-brush",
):
    """Run a lintian fixer on a tree.

    Args:
      local_tree: WorkingTree object
      basis_tree: Tree
      fixer: Fixer object to apply
      committer: Optional committer (name and email)
      update_changelog: Whether to add a new entry to the changelog
      compat_release: Minimum release that the package should be usable on
        (e.g. 'stable' or 'unstable')
      minimum_certainty: How certain the fixer should be
        about its changes.
      trust_package: Whether to run code from the package if necessary
      allow_reformatting: Whether to allow reformatting of changed files
      dirty_tracker: Optional object that can be used to tell if the tree
        has been changed.
      subpath: Path in tree to operate on
      net_access: Whether to allow accessing external services
      opinionated: Whether to be opinionated
      diligence: Level of diligence
    Returns:
      tuple with set of FixerResult, summary of the changes
    """
    if basis_tree is None:
        basis_tree = local_tree.basis_tree()

    changelog_path = os.path.join(subpath, "debian/changelog")

    try:
        with local_tree.get_file(changelog_path) as f:
            cl = Changelog(f, max_blocks=1)
    except NoSuchFile:
        raise NotDebianPackage(local_tree, subpath)
    package = cl.package
    if cl.distributions == "UNRELEASED":
        current_version = cl.version
    else:
        current_version = cl.version
        increment_version(current_version)
    if compat_release is None:
        compat_release = "sid"
    if minimum_certainty is None:
        minimum_certainty = DEFAULT_MINIMUM_CERTAINTY
    logger.debug('Running fixer %r', fixer)
    try:
        result = fixer.run(
            local_tree.abspath(subpath),
            package=package,
            current_version=current_version,
            compat_release=compat_release,
            minimum_certainty=minimum_certainty,
            trust_package=trust_package,
            allow_reformatting=allow_reformatting,
            net_access=net_access,
            opinionated=opinionated,
            diligence=diligence,
        )
    except BaseException:
        reset_tree(local_tree, basis_tree, subpath, dirty_tracker=dirty_tracker)
        raise
    if not certainty_sufficient(result.certainty, minimum_certainty):
        reset_tree(local_tree, basis_tree, subpath, dirty_tracker=dirty_tracker)
        raise NotCertainEnough(
            fixer, result.certainty, minimum_certainty,
            overridden_lintian_issues=result.overridden_lintian_issues)
    specific_files: Optional[List[str]]
    if dirty_tracker:
        relpaths = dirty_tracker.relpaths()
        # Sort paths so that directories get added before the files they
        # contain (on VCSes where it matters)
        local_tree.add(
            [
                p
                for p in sorted(relpaths)
                if local_tree.has_filename(p) and not local_tree.is_ignored(p)
            ]
        )
        specific_files = [p for p in relpaths if local_tree.is_versioned(p)]
        if not specific_files:
            raise NoChanges(
                fixer, "Script didn't make any changes",
                result.overridden_lintian_issues)
    else:
        local_tree.smart_add([local_tree.abspath(subpath)])
        specific_files = [subpath] if subpath else None

    if local_tree.supports_setting_file_ids():
        RenameMap.guess_renames(basis_tree, local_tree, dry_run=False)

    changes = list(
        local_tree.iter_changes(
            basis_tree,
            specific_files=specific_files,
            want_unversioned=False,
            require_versioned=True,
        )
    )

    if len(local_tree.get_parent_ids()) <= 1 and not changes:
        raise NoChanges(
            fixer, "Script didn't make any changes",
            result.overridden_lintian_issues)

    if not result.description:
        raise DescriptionMissing(fixer)

    lines = result.description.splitlines()
    summary = lines[0]
    details = list(itertools.takewhile(lambda line: line, lines[1:]))

    # If there are upstream changes in a non-native package, perhaps
    # export them to debian/patches
    if has_non_debian_changes(changes, subpath) and current_version.debian_revision:
        try:
            patch_name, specific_files = _upstream_changes_to_patch(
                local_tree,
                basis_tree,
                dirty_tracker,
                subpath,
                result.patch_name or fixer.name,
                result.description,
                timestamp=timestamp,
            )
        except BaseException:
            reset_tree(local_tree, basis_tree, subpath, dirty_tracker=dirty_tracker)
            raise

        summary = "Add patch %s: %s" % (patch_name, summary)

    if update_changelog is None:
        from .detect_gbp_dch import guess_update_changelog

        dch_guess = guess_update_changelog(
            local_tree, os.path.join(subpath, "debian"), cl)
        if dch_guess:
            update_changelog = dch_guess[0]
            _note_changelog_policy(update_changelog, dch_guess[1])
        else:
            # Assume we should update changelog
            update_changelog = True

    if update_changelog and only_changes_last_changelog_block(
        local_tree, basis_tree, changelog_path, changes
    ):
        # If the script only changed the last entry in the changelog,
        # don't update the changelog
        update_changelog = False

    if update_changelog:
        from .changelog import add_changelog_entry

        add_changelog_entry(local_tree, changelog_path, [summary] + details)
        if specific_files:
            specific_files.append(changelog_path)

    description = result.description + "\n"
    description += "\n"
    description += "Changes-By: %s\n" % changes_by
    for tag in result.fixed_lintian_tags:
        description += "Fixes: lintian: %s\n" % tag
        description += "See-also: https://lintian.debian.org/tags/%s.html\n" % tag

    if committer is None:
        committer = get_committer(local_tree)

    revid = local_tree.commit(
        description,
        allow_pointless=False,
        reporter=NullCommitReporter(),
        committer=committer,
        specific_files=specific_files,
    )
    result.revision_id = revid
    # TODO(jelmer): Support running sbuild & verify lintian warning is gone?
    return result, summary


class ManyResult(object):
    def __init__(self):
        self.success = []
        self.failed_fixers = {}
        self.formatting_unpreservable = {}
        self.overridden_lintian_issues = []

    def minimum_success_certainty(self) -> str:
        """Return the minimum certainty of any successfully made change."""
        return min_certainty(
            [
                r.certainty
                for r, unused_summary in self.success
                if r.certainty is not None
            ]
        )

    def __tuple__(self):
        return (self.success, self.failed_fixers)

    def __iter__(self):
        return iter(self.__tuple__())


def get_dirty_tracker(
    local_tree: WorkingTree, subpath: str = "", use_inotify: bool = None
):
    """Create a dirty tracker object."""
    if use_inotify is True:
        from breezy.dirty_tracker import DirtyTracker

        return DirtyTracker(local_tree, subpath)
    elif use_inotify is False:
        return None
    else:
        try:
            from breezy.dirty_tracker import DirtyTracker
        except ImportError:
            return None
        else:
            return DirtyTracker(local_tree, subpath)


def run_lintian_fixers(
    local_tree: WorkingTree,
    fixers: List[Fixer],
    update_changelog: bool = True,
    verbose: bool = False,
    committer: Optional[str] = None,
    compat_release: Optional[str] = None,
    minimum_certainty: Optional[str] = None,
    trust_package: bool = False,
    allow_reformatting: bool = False,
    use_inotify: Optional[bool] = None,
    subpath: str = "",
    net_access: bool = True,
    opinionated: Optional[bool] = None,
    diligence: int = 0,
):
    """Run a set of lintian fixers on a tree.

    Args:
      local_tree: WorkingTree object
      fixers: A set of Fixer objects
      update_changelog: Whether to add an entry to the changelog
      verbose: Whether to be verbose
      committer: Optional committer (name and email)
      compat_release: Minimum release that the package should be usable on
        (e.g. 'sid' or 'stretch')
      minimum_certainty: How certain the fixer should be
        about its changes.
      trust_package: Whether to run code from the package if necessary
      allow_reformatting: Whether to allow reformatting of changed files
      use_inotify: Use inotify to watch changes (significantly improves
        performance). Defaults to None (automatic)
      subpath: Subpath in the tree in which the package lives
      net_access: Whether to allow network access
      opinionated: Whether to be opinionated
      diligence: Level of diligence
    Returns:
      Tuple with two lists:
        1. list of tuples with (lintian-tag, certainty, description) of fixers
           that ran
        2. dictionary mapping fixer names for fixers that failed to run to the
           error that occurred
    """
    basis_tree = local_tree.basis_tree()
    check_clean_tree(local_tree, basis_tree=basis_tree, subpath=subpath)
    fixers = list(fixers)
    dirty_tracker = get_dirty_tracker(
        local_tree, subpath=subpath, use_inotify=use_inotify
    )
    ret = ManyResult()
    with trange(len(fixers), leave=False, disable=None) as t:
        for i, fixer in enumerate(fixers):
            t.set_description("Running fixer %s" % fixer)
            t.update()
            start = time.time()
            if dirty_tracker:
                dirty_tracker.mark_clean()
            try:
                result, summary = run_lintian_fixer(
                    local_tree,
                    fixer,
                    update_changelog=update_changelog,
                    committer=committer,
                    compat_release=compat_release,
                    minimum_certainty=minimum_certainty,
                    trust_package=trust_package,
                    allow_reformatting=allow_reformatting,
                    dirty_tracker=dirty_tracker,
                    subpath=subpath,
                    net_access=net_access,
                    opinionated=opinionated,
                    diligence=diligence,
                    basis_tree=basis_tree,
                )
            except FormattingUnpreservable as e:
                ret.formatting_unpreservable[fixer.name] = e.path
                if verbose:
                    logging.info(
                        "Fixer %r was unable to preserve formatting of %s.",
                        fixer.name,
                        e.path,
                    )
            except FixerFailed as e:
                ret.failed_fixers[fixer.name] = e
                if verbose:
                    logging.info("Fixer %r failed to run.", fixer.name)
                    sys.stderr.write(str(e))
            except MemoryError as e:
                ret.failed_fixers[fixer.name] = e
                if verbose:
                    logging.info("Run out of memory while running fixer %r.", fixer.name)
            except NotCertainEnough as e:
                if verbose:
                    logging.info(
                        "Fixer %r made changes but not high enough "
                        "certainty (was %r, needed %r). (took: %.2fs)",
                        fixer.name,
                        e.certainty,
                        e.minimum_certainty,
                        time.time() - start,
                    )
            except FailedPatchManipulation as e:
                if verbose:
                    logging.info("Unable to manipulate upstream patches: %s", e.args[2])
                ret.failed_fixers[fixer.name] = e
            except NoChanges as e:
                if verbose:
                    logging.info(
                        "Fixer %r made no changes. (took: %.2fs)",
                        fixer.name,
                        time.time() - start,
                    )
                ret.overridden_lintian_issues.extend(
                    e.overridden_lintian_issues)
            else:
                if verbose:
                    logging.info(
                        "Fixer %r made changes. (took %.2fs)",
                        fixer.name,
                        time.time() - start,
                    )
                ret.success.append((result, summary))
                basis_tree = local_tree.basis_tree()
    return ret


def certainty_to_confidence(certainty: Optional[str]) -> Optional[int]:
    if certainty in ("unknown", None):
        return None
    return SUPPORTED_CERTAINTIES.index(certainty)


def confidence_to_certainty(confidence: Optional[int]) -> str:
    if confidence is None:
        return "unknown"
    try:
        return SUPPORTED_CERTAINTIES[confidence] or "unknown"
    except IndexError:
        raise ValueError(confidence)


def min_certainty(certainties: Sequence[str]) -> str:
    return confidence_to_certainty(
        max([SUPPORTED_CERTAINTIES.index(c) for c in certainties] + [0])
    )


def control_files_in_root(tree: Tree, subpath: str) -> bool:
    debian_path = os.path.join(subpath, "debian")
    if tree.has_filename(debian_path):
        return False
    control_path = os.path.join(subpath, "control")
    if tree.has_filename(control_path):
        return True
    if tree.has_filename(control_path + ".in"):
        return True
    return False


def is_debcargo_package(tree: Tree, subpath: str) -> bool:
    control_path = os.path.join(subpath, "debian", "debcargo.toml")
    return tree.has_filename(control_path)


def control_file_present(tree: Tree, subpath: str) -> bool:
    """Check whether there are any control files present in a tree.

    Args:
      tree: Tree to check
      subpath: subpath to check
    Returns:
      whether control file is present
    """
    for name in ["debian/control", "debian/control.in", "control",
                 "control.in", "debian/debcargo.toml"]:
        name = os.path.join(subpath, name)
        if tree.has_filename(name):
            return True
    return False



#!/usr/bin/python3
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with watch files."""

from dataclasses import dataclass
import json
import os
import re
from typing import Set, Optional
from urllib.parse import urlparse, urlunparse
import urllib.error
from urllib.request import urlopen, Request

from debmutate.watch import Watch

from . import (
    USER_AGENT,
    DEFAULT_URLLIB_TIMEOUT,
)


@dataclass
class WatchCandidate:

    watch: Watch
    site: str
    certainty: Optional[str]
    preference: Optional[int]


def candidates_from_setup_py(path, good_upstream_versions: Set[str], net_access=False):
    certainty = "likely"
    from distutils.core import run_setup

    try:
        result = run_setup(os.path.abspath(path), stop_after="init")
    except BaseException:
        import traceback

        traceback.print_exc()
        return
    project = result.get_name()  # type: ignore
    version = result.get_version()  # type: ignore
    if not project:
        return
    current_version_filenames = None
    if net_access:
        json_url = "https://pypi.python.org/pypi/%s/json" % project
        headers = {"User-Agent": USER_AGENT}
        try:
            response = urlopen(
                Request(json_url, headers=headers), timeout=DEFAULT_URLLIB_TIMEOUT
            )
        except urllib.error.HTTPError as e:
            if e.status == 404:
                return
            raise
        pypi_data = json.load(response)
        if version in pypi_data["releases"]:
            release = pypi_data["releases"][version]
            current_version_filenames = [
                (d["filename"], d["has_sig"])
                for d in release
                if d["packagetype"] == "sdist"
            ]
    filename_regex = r"%(project)s-(.+)\.(?:zip|tgz|tbz|txz|(?:tar\.(?:gz|bz2|xz)))" % {
        "project": project
    }
    opts = []
    # TODO(jelmer): Set uversionmangle?
    # opts.append('uversionmangle=s/(rc|a|b|c)/~$1/')
    if current_version_filenames:
        for (fn, has_sig) in current_version_filenames:
            if re.match(filename_regex, fn):
                certainty = "certain"
                if has_sig:
                    opts.append("pgpsigurlmangle=s/$/.asc/")
    url = r"https://pypi.debian.net/%(project)s/%(fname_regex)s" % {
        "project": project,
        "fname_regex": filename_regex,
    }
    # TODO(jelmer): Add pgpsigurlmangle if has_sig==True
    w = Watch(url, opts=opts)
    yield WatchCandidate(w, "pypi", certainty=certainty, preference=1)


def candidates_from_upstream_metadata(
    path: str, good_upstream_versions: Set[str], net_access: bool = False
):
    try:
        with open(path, "r") as f:
            inp = f.read()
    except FileNotFoundError:
        pass
    else:
        import ruamel.yaml

        code = ruamel.yaml.round_trip_load(inp, preserve_quotes=True)

        try:
            parsed_url = urlparse(code["Repository"])
        except KeyError:
            pass
        else:
            if parsed_url.hostname == "github.com":
                yield from guess_github_watch_entry(
                    parsed_url, good_upstream_versions, net_access=net_access
                )

        archive = code.get('Archive')
        if archive == 'CRAN':
            yield from guess_cran_watch_entry(code['Name'])


def guess_cran_watch_entry(name):
    w = Watch(r'https://cran.r-project.org/src/contrib/%s_([-\d.]*)\.tar\.gz' % name)
    yield WatchCandidate(w, "cran", certainty="likely", preference=0)


def guess_github_watch_entry(parsed_url, good_upstream_versions, net_access=False):
    from breezy.branch import Branch
    import re

    if not net_access:
        return
    branch = Branch.open(urlunparse(parsed_url))
    tags = branch.tags.get_tag_dict()
    POSSIBLE_PATTERNS = [r"v(\d\S+)", r"(\d\S+)"]
    version_pattern = None
    # TODO(jelmer): Maybe use releases API instead?
    # TODO(jelmer): Automatically added mangling for
    # e.g. rc and beta
    uversionmangle = []
    for name in sorted(tags, reverse=True):
        for pattern in POSSIBLE_PATTERNS:
            m = re.match(pattern, name)
            if not m:
                continue
            if m.group(1) in good_upstream_versions:
                version_pattern = pattern
                break
        if version_pattern:
            break
    else:
        return
    (username, project) = parsed_url.path.strip("/").split("/")
    if project.endswith(".git"):
        project = project[:-4]
    download_url = "https://github.com/%(user)s/%(project)s/tags" % {
        "user": username,
        "project": project,
    }
    matching_pattern = r".*\/%s\.tar\.gz" % version_pattern
    opts = []
    opts.append(
        r"filenamemangle=s/%(pattern)s/%(project)s-$1\.tar\.gz/"
        % {"pattern": matching_pattern, "project": project}
    )
    if uversionmangle:
        opts.append(r"uversionmangle=" + ";".join(uversionmangle))
    # TODO(jelmer): Check for GPG
    # opts.append(
    #    r'pgpsigurlmangle='
    #    r's/archive\/%s\.tar\.gz/releases\/download\/$1\/$1\.tar\.gz\.asc/' %
    #    version_pattern)
    w = Watch(download_url, matching_pattern, opts=opts)
    yield WatchCandidate(w, "github", certainty="certain", preference=0)


def candidates_from_hackage(package, good_upstream_versions, net_access=False):
    if not net_access:
        return
    url = "https://hackage.haskell.org/package/%s/preferred" % package
    headers = {"User-Agent": USER_AGENT, "Accept": "application/json"}
    try:
        response = urlopen(
            Request(url, headers=headers), timeout=DEFAULT_URLLIB_TIMEOUT
        )
    except urllib.error.HTTPError as e:
        if e.status == 404:
            return
        raise
    versions = json.load(response)
    for version in versions["normal-version"]:
        if version in good_upstream_versions:
            break
    else:
        return
    download_url = "https://hackage.haskell.org/package/" + package
    matching_pattern = r".*/%s-(.*).tar.gz" % package
    w = Watch(download_url, matching_pattern)
    yield WatchCandidate(w, "hackage", certainty="certain", preference=1)


def fix_old_github_patterns(updater):
    for w in getattr(updater.watch_file, "entries", []):
        parsed_url = urlparse(w.url)

        # only applies to github.com
        if parsed_url.netloc != "github.com":
            continue

        parts = parsed_url.path.strip('/').split('/')
        if len(parts) >= 3 and parts[2] in ('tags', 'releases'):
            pass
        elif len(parts) >= 2 and parts[0] == '.*' and parts[1] == 'archive':
            pass
        else:
            continue

        parts = w.matching_pattern.split('/')
        if len(parts) > 2 and parts[-2] == 'archive':
            parts.insert(-1, 'refs/tags')
        w.matching_pattern = '/'.join(parts)


def fix_watch_issues(updater):
    fix_old_github_patterns(updater)



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.salsa."""

from breezy.tests import (
    TestCase,
)
from lintian_brush.salsa import (
    guess_repository_url,
    determine_browser_url,
    salsa_url_from_alioth_url,
)


class GuessRepositoryURLTests(TestCase):
    def test_unknown(self):
        self.assertIs(
            None, guess_repository_url("blah", "unknown-team@lists.alioth.debian.org")
        )

    def test_individual(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/lintian-brush.git",
            guess_repository_url("lintian-brush", "jelmer@debian.org"),
        )

    def test_team(self):
        self.assertEqual(
            "https://salsa.debian.org/js-team/node-blah.git",
            guess_repository_url(
                "node-blah", "pkg-javascript-devel@lists.alioth.debian.org"
            ),
        )


class DetermineBrowserUrlTests(TestCase):
    def test_browser_url(self):
        self.assertEqual(
            "https://salsa.debian.org/js-team/node-blah",
            determine_browser_url("https://salsa.debian.org/js-team/node-blah.git"),
        )
        self.assertEqual(
            "https://salsa.debian.org/js-team/node-blah",
            determine_browser_url("https://salsa.debian.org/js-team/node-blah"),
        )

    def test_branch(self):
        self.assertEqual(
            "https://salsa.debian.org/js-team/node-blah/tree/debian",
            determine_browser_url(
                "https://salsa.debian.org/js-team/node-blah -b debian"
            ),
        )

    def test_subpath(self):
        self.assertEqual(
            "https://salsa.debian.org/js-team/node-blah/tree/debian/foo",
            determine_browser_url(
                "https://salsa.debian.org/js-team/node-blah -b debian [foo]"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/js-team/node-blah/tree/HEAD/foo",
            determine_browser_url("https://salsa.debian.org/js-team/node-blah [foo]"),
        )


class SalsaUrlFromAliothUrlTests(TestCase):
    def test_none(self):
        self.assertIs(None, salsa_url_from_alioth_url(None, None))

    def test_mismatch(self):
        self.assertIs(
            None, salsa_url_from_alioth_url("bzr", "https://code.launchpad.net/blah")
        )

    def test_perl(self):
        self.assertEqual(
            "https://salsa.debian.org/perl-team/modules/packages/libbla.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/pkg-perl/trunk/libbla"
            ),
        )

    def test_git(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/python-bla.git",
            salsa_url_from_alioth_url(
                "git", "http://anonscm.debian.org/git/users/jelmer/python-bla"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/jelmer/python-bla.git",
            salsa_url_from_alioth_url(
                "git", "http://anonscm.debian.org/users/jelmer/python-bla"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/go-team/golang-example-blah-blah.git",
            salsa_url_from_alioth_url(
                "git", "http://anonscm.debian.org/pkg-go/golang-example-blah-blah"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/lua-team/blah.git",
            salsa_url_from_alioth_url(
                "git", "https://alioth.debian.org/anonscm/git/pkg-lua/blah"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/science-team/ros-geometry.git",
            salsa_url_from_alioth_url(
                "git",
                "https://anonscm.debian.org/cgit/debian-science/" "ros-geometry.git",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/nagios-team/pkg-check-multi.git",
            salsa_url_from_alioth_url(
                "git",
                "https://anonscm.debian.org/git/pkg-nagios/" "pkg-check-multi.git",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/perl-team/modules/packages/"
            "libgstream-interfaces-perl.git",
            salsa_url_from_alioth_url(
                "git",
                "git://git.debian.org/pkg-perl/packages/"
                "libgstream-interfaces-perl.git",
            ),
        )
        # TODO(jelmer): This should actually be
        # https://salsa.debian.org/qt-kde-team/extras/plasma-widget-menubar.git
        self.assertEqual(
            "https://salsa.debian.org/qt-kde-team/"
            "kde-extras/plasma-widget-menubar.git",
            salsa_url_from_alioth_url(
                "git",
                "git://anonscm.debian.org/pkg-kde/kde-extras/"
                "plasma-widget-menubar.git",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/fonts-team/fonts-beteckna.git",
            salsa_url_from_alioth_url(
                "git", "git://anonscm.debian.org/pkg-fonts/fonts-beteckna.git"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/brlink/xwit.git",
            salsa_url_from_alioth_url(
                "git", "git://anonscm.debian.org/users/brlink/xwit.git"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/qt-kde-team/kde/kruler.git",
            salsa_url_from_alioth_url(
                "git", "https://anonscm.debian.org/git/pkg-kde/applications/" "kruler"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/3dprinting-team/arduino-mighty.git",
            salsa_url_from_alioth_url(
                "git",
                "https://anonscm.debian.org/git/3dprinter/packages/" "arduino-mighty",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/emacsen-team/lua-mode.git",
            salsa_url_from_alioth_url(
                "git", "https://anonscm.debian.org/git/pkg-emacsen/pkg/lua-mode"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/debian-astro-team/astromatic.git",
            salsa_url_from_alioth_url(
                "git",
                "http://anonscm.debian.org/cgit/debian-astro/packages/"
                "astromatic.git",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/debichem-team/bkchem.git",
            salsa_url_from_alioth_url(
                "git", "https://anonscm.debian.org/git/debichem/packages/bkchem.git"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/3dprinting-team/yagv.git",
            salsa_url_from_alioth_url(
                "git", "https://anonscm.debian.org/cgit/3dprinter/packages/yagv.git"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/fonts-team/fonts-pagul.git",
            salsa_url_from_alioth_url(
                "git", "git://anonscm.debian.org/debian-in/fonts-pagul.git"
            ),
        )

    def test_svn(self):
        self.assertEqual(
            "https://salsa.debian.org/multimedia-team/ezstream.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/svn/pkg-icecast/ezstream/trunk"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/bsd-team/freebsd-buildutils.git",
            salsa_url_from_alioth_url(
                "svn", "svn://anonscm.debian.org/glibc-bsd/trunk/freebsd-buildutils/"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/nvidia-team/nvclock.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/pkg-nvidia/packages/nvclock/trunk"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/llvm-team/llvm.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/svn/pkg-llvm/llvm/trunk/"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/xfce-team/xfswitch-plugin.git",
            salsa_url_from_alioth_url(
                "svn",
                "svn://anonscm.debian.org/pkg-xfce/goodies/trunk/" "xfswitch-plugin/",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/python-team/applications/pypar2.git",
            salsa_url_from_alioth_url(
                "svn", "svn://anonscm.debian.org/python-apps/packages/pypar2/trunk/"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/xml-sgml-team/docbook-slides-demo.git",
            salsa_url_from_alioth_url(
                "svn",
                "svn://svn.debian.org/debian-xml-sgml/packages/"
                "docbook-slides-demo/trunk/",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/fonts-team/ttf-inconsolata.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/svn/pkg-fonts/packages/ttf-inconsolata"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/qt-kde-team/kde3libs.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/svn/pkg-kde/trunk/packages/kde3libs"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/python-team/applications/" "upnp-inspector.git",
            salsa_url_from_alioth_url(
                "svn",
                "svn://svn.debian.org/python-apps/packages/upnp-inspector/" "trunk/",
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/python-team/applications/hotssh.git",
            salsa_url_from_alioth_url(
                "svn", "svn://anonscm.debian.org/python-apps/packages/hotssh/trunk/"
            ),
        )
        self.assertEqual(
            "https://salsa.debian.org/debichem-team/drawxtl.git",
            salsa_url_from_alioth_url(
                "svn", "svn://svn.debian.org/svn/debichem/unstable/drawxtl"
            ),
        )



#!/usr/bin/python3

import sys
from lintian_brush.fixer import (
    control,
    report_result,
    meets_minimum_certainty,
    is_debcargo_package,
    LintianIssue,
    )

require_root = "no"
CERTAINTY = "possible"

if not meets_minimum_certainty(CERTAINTY):
    sys.exit(0)

if is_debcargo_package():
    sys.exit(2)

with control as updater:
    if "Rules-Requires-Root" not in updater.source:
        issue = LintianIssue(updater.source, 'silent-on-rules-requiring-root')
        if issue.should_fix():
            # TODO: add some heuristics to set require_root = "yes" in common
            # cases, like `debian/rules binary` chown(1)'ing stuff
            updater.source["Rules-Requires-Root"] = require_root
            issue.report_fixed()

report_result(
    "Set Rules-Requires-Root: %s." % require_root,
    certainty=CERTAINTY)



#!/usr/bin/python3

import sys

from lintian_brush.fixer import control, report_result, LintianIssue

removed = []

try:
    with control as updater:
        for binary in updater.binaries:
            for field, value in binary.items():
                if updater.source.get(field) == value:
                    del binary[field]
                    removed.append((binary['Package'], field, value))
                    issue = LintianIssue(
                        updater.source,
                        'installable-field-mirrors-source',
                        info='field "%s" in package %s' % (
                            field, binary['Package']))
                    issue.report_fixed()
except FileNotFoundError:
    sys.exit(0)


report_result(
    'Remove %s that duplicate%s source.' % (
        ', '.join(
            ['%s on %s' % (field, package)
             for (package, field, value) in removed]),
        's' if len(removed) == 1 else ''))



#!/usr/bin/python3

from setuptools import setup

setup(
    name='dulwich',
    description='Python Git Library',
    version="0.1.1",
    author="Jelmer Vernooij",
    author_email="jelmer@jelmer.uk",
    url="https://www.dulwich.io/",
)



#!/usr/bin/python3
import os
import sys

from debmutate.watch import WatchFile

from lintian_brush import (
    certainty_to_confidence,
    )
from lintian_brush.fixer import (
    current_package_version,
    net_access_allowed,
    report_result,
    package_is_native,
    LintianIssue,
    )
from lintian_brush.watch import (
    candidates_from_setup_py,
    candidates_from_upstream_metadata,
    )


if os.path.exists('debian/watch') or package_is_native():
    # Nothing to do here..
    sys.exit(0)

issue = LintianIssue('source', 'debian-watch-file-is-missing', info='')

if not issue.should_fix():
    sys.exit(0)

candidates = []

good_upstream_versions = {current_package_version().upstream_version}

if os.path.exists('setup.py'):
    candidates.extend(candidates_from_setup_py(
        'setup.py', good_upstream_versions,
        net_access=net_access_allowed()))

if os.path.exists('debian/upstream/metadata'):
    candidates.extend(candidates_from_upstream_metadata(
        'debian/upstream/metadata', good_upstream_versions,
        net_access=net_access_allowed()))


def candidate_key(candidate):
    return (
        certainty_to_confidence(candidate.certainty),
        candidate.preference)


candidates.sort(key=candidate_key)

# TODO(jelmer): parse cabal file and call candidates_from_hackage

if not candidates:
    sys.exit(0)

winner = candidates[0]

wf = WatchFile()
wf.entries.append(winner.watch)

with open('debian/watch', 'w') as f:
    wf.dump(f)
    issue.report_fixed()

report_result(
    "Add debian/watch file, using %s." % winner.site, certainty=winner.certainty)



#!/usr/bin/python3

from debmutate.changelog import ChangelogEditor, changes_sections
from lintian_brush.fixer import report_result

with ChangelogEditor() as updater:
    block = updater.changelog[0]
    to_delete = set()
    if block.distributions == 'UNRELEASED':
        seen = {}
        for (author, section_linenos, section_contents) in changes_sections(
                block.changes()):
            keep_section = False
            for entry in section_contents:
                change = ''.join([line for (lineno, line) in entry])
                if (author, change) in seen:
                    to_delete.update([lineno for (lineno, line) in entry])
                else:
                    seen[(author, change)] = entry
                    keep_section = True
            if not keep_section:
                to_delete.update(section_linenos)
    for lineno in sorted(to_delete, reverse=True):
        del block._changes[lineno]


report_result('Remove duplicate line from changelog.')



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Section handling."""

import os
import re


from .lintian import LINTIAN_DATA_PATH


NAME_SECTION_MAPPINGS_PATH = os.path.join(LINTIAN_DATA_PATH, "fields/name_section_mappings")


def get_name_section_mappings():
    regexes = []

    with open(NAME_SECTION_MAPPINGS_PATH, "r") as f:
        for line in f:
            if line.startswith('#'):
                continue
            if not line.strip():
                continue
            (regex, section) = line.split("=>")
            regexes.append((re.compile(regex.strip()), section.strip()))
    return regexes


def find_expected_section(regexes, name):
    for regex, section in regexes:
        if regex.match(name):
            return section
    return None



#!/usr/bin/python3

from lintian_brush.fixer import report_result, LintianIssue

import os


def is_empty(path):
    with open(path, 'rb') as f:
        for line in f.readlines():
            line = line.strip()
            if not line:
                continue
            if line.startswith(b'#'):
                continue
            if line.startswith(b'set '):
                continue
            if line.startswith(b'exit '):
                continue
            return False
    return True


MAINTAINER_SCRIPTS = ['prerm', 'postinst', 'preinst', 'postrm']
removed = []

for entry in os.scandir('debian'):
    if entry.name in MAINTAINER_SCRIPTS:
        script = entry.name
        package = None
    elif '.' not in entry.name:
        continue
    else:
        parts = entry.name.rsplit('.', 1)
        if len(parts) != 2:
            continue
        package = parts[0]
        script = parts[1]
        if script not in MAINTAINER_SCRIPTS:
            continue
    if is_empty(entry.path):
        issue = LintianIssue(package, 'maintainer-script-empty', script)
        if issue.should_fix():
            removed.append((package, script))
            os.unlink(entry.path)
            issue.report_fixed()

report_result(
    'Remove empty maintainer scripts: ' +
    ', '.join('%s (%s)' % x for x in removed))



#!/usr/bin/python3

from lintian_brush.fixer import control, report_result, fixed_lintian_tag

FIXABLE_HOSTS = [
    'gitlab.com', 'github.com', 'salsa.debian.org',
    'gitorious.org']


with control as updater:
    vcs_git = updater.source.get("Vcs-Git")
    if vcs_git and ':' in vcs_git:
        (netloc, path) = vcs_git.split(':', 1)
        if netloc.startswith('git@'):
            netloc = netloc[4:]
        if netloc in FIXABLE_HOSTS:
            fixed_lintian_tag(
                updater.source, 'vcs-field-uses-not-recommended-uri-format',
                info='vcs-git %s' % updater.source['Vcs-Git'])
            updater.source["Vcs-Git"] = 'https://%s/%s' % (netloc, path)


report_result("Use recommended URI format in Vcs header.")



#!/usr/bin/python3

from urllib.parse import urlparse

from lintian_brush.fixer import (
    control,
    meets_minimum_certainty,
    report_result,
    trust_package,
    LintianIssue,
    )
from upstream_ontologist.guess import guess_upstream_metadata_items, known_bad_guess


def guess_homepage():
    for datum in guess_upstream_metadata_items(
            '.', trust_package=trust_package()):
        if datum.field != 'Homepage':
            continue
        if not meets_minimum_certainty(datum.certainty):
            continue
        if known_bad_guess(datum):
            continue
        return datum


with control as updater:
    if 'Homepage' not in updater.source:
        datum = guess_homepage()
        issue = LintianIssue('source', 'no-homepage-field')
        if datum and issue.should_fix():
            updater.source["Homepage"] = datum.value
            issue.report_fixed()
            report_result('Fill in Homepage field.', certainty=datum.certainty)
    else:
        hostname = urlparse(updater.source['Homepage']).hostname
        if hostname == 'pypi.org':
            issue = LintianIssue('source', 'pypi-homepage', updater.source['Homepage'])
        elif hostname == 'rubygems.org':
            issue = LintianIssue('source', 'pypi-homepage', updater.source['Homepage'])
        else:
            issue = None

        if issue:
            datum = guess_homepage()
            if issue.should_fix() and datum:
                updater.source["Homepage"] = datum.value
                issue.report_fixed()
                report_result('Avoid %s in Homepage field.' % hostname, certainty=datum.certainty)



#!/usr/bin/python3

import sys
from debmutate.control import drop_dependency
from lintian_brush.fixer import control, report_result, fixed_lintian_tag

try:
    with open('debian/rules', 'rb') as f:
        for line in f:
            if b'/usr/share/cdbs/' in line:
                uses_cdbs = True
                break
        else:
            uses_cdbs = False
except FileNotFoundError:
    # Unsure whether it actually needs cdbs
    sys.exit(2)

if not uses_cdbs:
    with control as updater:
        new_depends = drop_dependency(
            updater.source.get("Build-Depends", ""), "cdbs")
        if new_depends != updater.source['Build-Depends']:
            fixed_lintian_tag(
                updater.source, 'unused-build-dependency-on-cdbs', '')
            updater.source["Build-Depends"] = new_depends
        if not updater.source["Build-Depends"]:
            del updater.source["Build-Depends"]

report_result("Drop unused build-dependency on cdbs.")



#!/usr/bin/python3

from datetime import datetime, timedelta
import os
import email.utils
import sys

from debmutate.changelog import ChangelogEditor
from debmutate.debhelper import MaintscriptEditor
from lintian_brush.debhelper import drop_obsolete_maintscript_entries
from lintian_brush.fixer import report_result, upgrade_release, warn


# If there is no information from the upgrade release, default to 5 years.
DEFAULT_AGE_THRESHOLD_DAYS = 5 * 365


maintscripts = []
for entry in os.scandir('debian'):
    if not (entry.name == "maintscript" or entry.name.endswith(".maintscript")):
        continue
    maintscripts.append(entry.name)


# Determine the date for which versions created then should still be supported.
# This is a little bit tricky since versions uploaded at a particular date
# may not have made it into the release then.
from distro_info import DebianDistroInfo  # noqa: E402
try:
    [release] = [
        r for r in DebianDistroInfo().get_all('object')
        if r.codename.lower() == upgrade_release()]
except ValueError:
    date_threshold = None
else:
    date_threshold = release.release

if date_threshold is None:
    # Release has not yet or will never be released
    # Default to 5 years
    date_threshold = (datetime.now() - timedelta(days=DEFAULT_AGE_THRESHOLD_DAYS)).date()


cl_dates = []
with ChangelogEditor() as cl:
    for block in cl:
        try:
            dt = email.utils.parsedate_to_datetime(block.date)
        except (TypeError, ValueError):
            warn('Invalid date %r for %s' % (block.date, block.version))
            # parsedate_to_datetime is buggy and raises a TypeError
            # when the date is invalid.
            # We can't reliably check anymore :(
            sys.exit(2)
        cl_dates.append((block.version, dt))


def is_well_past(version):
    for (cl_version, cl_dt) in cl_dates:
        if cl_version <= version and cl_dt.date() > date_threshold:
            return False
    return True


total_entries = 0
ret = []
for name in maintscripts:
    with MaintscriptEditor(os.path.join('debian', name)) as editor:
        removed = drop_obsolete_maintscript_entries(
            editor, lambda p, v: is_well_past(v))
        if removed:
            ret.append((os.path.join('debian', name), removed))
            total_entries += len(removed)

if total_entries == 1:
    report_result('Remove %d obsolete maintscript entry.' % total_entries)
else:
    report_result('Remove %d obsolete maintscript entries in %d files.' %
                  (total_entries, len(ret)))



#!/usr/bin/python3

from debian.copyright import (
    LicenseParagraph,
    NotMachineReadableError,
    )
from debmutate.copyright import CopyrightEditor
from lintian_brush.fixer import report_result, fixed_lintian_tag


def fix_header_license_references(copyright):
    if not copyright.header.license:
        return
    if not copyright.header.license.text:
        return
    used_licenses = set()
    seen_licenses = set()
    for files_paragraph in copyright.all_files_paragraphs():
        if not files_paragraph.license:
            continue
        used_licenses.add(files_paragraph.license.synopsis)
        if files_paragraph.license.text:
            seen_licenses.add(files_paragraph.license.synopsis)
    for license_paragraph in copyright.all_license_paragraphs():
        seen_licenses.add(license_paragraph.license.synopsis)
    for missing in used_licenses - seen_licenses:
        if copyright.header.license.synopsis == missing:
            copyright.add_license_paragraph(
                LicenseParagraph.create(
                    copyright.header.license))
    fixed_lintian_tag(
        'source',
        'dep5-file-paragraph-references-header-paragraph',
        '%s (line XX)' % (copyright.header.license.synopsis, ))
    return copyright.header.license


try:
    with CopyrightEditor() as updater:
        license = fix_header_license_references(updater.copyright)
except (FileNotFoundError, NotMachineReadableError):
    pass
else:
    if license:
        report_result(
            'Add missing license paragraph for %s' % license.synopsis)



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for applying multi-arch hints."""

import contextlib
import json
import logging
import os
import re
import sys
import time

from urllib.error import HTTPError
from urllib.request import urlopen, Request

from breezy.workspace import (
    WorkspaceDirty,
    check_clean_tree,
    )

from lintian_brush import (
    Fixer,
    NoChanges,
    NotDebianPackage,
    FixerResult,
    min_certainty,
    USER_AGENT,
    SUPPORTED_CERTAINTIES,
    DEFAULT_URLLIB_TIMEOUT,
    certainty_sufficient,
    get_committer,
    get_dirty_tracker,
    run_lintian_fixer,
    version_string,
    control_files_in_root,
    control_file_present,
    is_debcargo_package,
)
from debmutate.control import (
    ControlEditor,
    format_relations,
    parse_relations,
)
from debmutate.reformatting import GeneratedFile, FormattingUnpreservable


DEFAULT_VALUE_MULTIARCH_HINT = 50
MULTIARCH_HINTS_VALUE = {
    "ma-foreign": 20,
    "file-conflict": 50,
    "ma-foreign-library": 20,
    "dep-any": 20,
    "ma-same": 20,
    "arch-all": 20,
}


def calculate_value(hints):
    return sum(map(MULTIARCH_HINTS_VALUE.__getitem__, hints)) + (
        DEFAULT_VALUE_MULTIARCH_HINT
    )


MULTIARCH_HINTS_URL = "https://dedup.debian.net/static/multiarch-hints.yaml.xz"


def parse_multiarch_hints(f):
    """Parse a multi-arch hints file.

    Args:
      f: File-like object to read from
    Returns:
      dictionary mapping binary package names to lists of hints
    """
    from ruamel.yaml import YAML

    yaml = YAML(typ="safe")
    data = yaml.load(f)
    if data.get("format") != "multiarch-hints-1.0":
        raise ValueError("invalid file format: %r" % data.get("format"))
    return data["hints"]


def multiarch_hints_by_binary(hints):
    ret = {}
    for entry in hints:
        ret.setdefault(entry["binary"], []).append(entry)
    return ret


def multiarch_hints_by_source(hints):
    ret = {}
    for entry in hints:
        if "source" not in entry:
            continue
        ret.setdefault(entry["source"], []).append(entry)
    return ret


@contextlib.contextmanager
def cache_download_multiarch_hints(url=MULTIARCH_HINTS_URL):
    """Load multi-arch hints from a URL, but use cached version if available."""
    cache_home = os.environ.get("XDG_CACHE_HOME")
    if not cache_home:
        cache_home = os.path.expanduser("~/.cache")
    cache_dir = os.path.join(cache_home, "lintian-brush")
    try:
        os.makedirs(cache_dir, exist_ok=True)
    except PermissionError:
        local_hints_path = None
        logging.warning("Unable to create %s; not caching.", cache_dir)
    else:
        local_hints_path = os.path.join(cache_dir, "multiarch-hints.yml")
    try:
        last_modified = os.path.getmtime(local_hints_path)
    except FileNotFoundError:
        last_modified = None
    try:
        with download_multiarch_hints(url=url, since=last_modified) as f:
            if local_hints_path is None:
                yield f
                return
            logging.info("Downloading new version of multi-arch hints.")
            with open(local_hints_path, "wb") as c:
                c.writelines(f)
    except HTTPError as e:
        if e.status != 304:
            raise
    yield open(local_hints_path, "rb")


@contextlib.contextmanager
def download_multiarch_hints(url=MULTIARCH_HINTS_URL, since: int = None):
    """Load multi-arch hints from a URL.

    Args:
      url: URL to read from
      since: Last modified timestamp
    Returns:
      multi-arch hints file
    """
    headers = {"User-Agent": USER_AGENT}
    if since is not None:
        headers["If-Modified-Since"] = time.strftime(
            "%a, %d %b %Y %H:%M:%S GMT", time.gmtime(since)
        )

    with urlopen(Request(url, headers=headers), timeout=DEFAULT_URLLIB_TIMEOUT) as f:
        if url.endswith(".xz"):
            import lzma

            # It would be nicer if there was a content-type, but there isn't
            # :-(
            f = lzma.LZMAFile(f)
        yield f


def apply_hint_ma_foreign(binary, hint):
    if binary.get("Multi-Arch") != "foreign":
        binary["Multi-Arch"] = "foreign"
        return "Add Multi-Arch: foreign."


def apply_hint_ma_foreign_lib(binary, hint):
    if binary.get("Multi-Arch") == "foreign":
        del binary["Multi-Arch"]
        return "Drop Multi-Arch: foreign."


def apply_hint_file_conflict(binary, hint):
    if binary.get("Multi-Arch") == "same":
        del binary["Multi-Arch"]
        return "Drop Multi-Arch: same."


def apply_hint_dep_any(binary, hint):
    m = re.match(
        "(.*) could have its dependency on (.*) annotated with :any",
        hint["description"],
    )
    if not m or m.group(1) != binary["Package"]:
        raise ValueError("unable to parse hint description: %r" % hint["description"])
    dep = m.group(2)
    if "Depends" not in binary:
        return
    changed = False
    relations = parse_relations(binary["Depends"])
    for entry in relations:
        (head_whitespace, relation, tail_whitespace) = entry
        if not isinstance(relation, str):  # formatting
            for r in relation:
                if r.name == dep and r.archqual != "any":
                    r.archqual = "any"
                    changed = True
    if not changed:
        return
    binary["Depends"] = format_relations(relations)
    return "Add :any qualifier for %s dependency." % dep


def apply_hint_ma_same(binary, hint):
    if binary.get("Multi-Arch") == "same":
        return
    binary["Multi-Arch"] = "same"
    return "Add Multi-Arch: same."


def apply_hint_arch_all(binary, hint):
    if binary["Architecture"] == "all":
        return
    binary["Architecture"] = "all"
    return "Make package Architecture: all."


class MultiArchHintApplier(object):
    def __init__(self, kind, fn, certainty):
        self.kind = kind
        self.fn = fn
        self.certainty = certainty


class MultiArchFixerResult(FixerResult):
    def __init__(self, description, certainty, changes):
        super(MultiArchFixerResult, self).__init__(
            description=description, certainty=certainty
        )
        self.changes = changes


def apply_multiarch_hints(hints, minimum_certainty="certain"):
    changes = []
    appliers = {applier.kind: applier for applier in APPLIERS}

    with ControlEditor() as editor:
        for binary in editor.binaries:
            for hint in hints.get(binary["Package"], []):
                kind = hint["link"].rsplit("#", 1)[1]
                applier = appliers[kind]
                if not certainty_sufficient(applier.certainty, minimum_certainty):
                    continue
                description = applier.fn(binary, hint)
                if description:
                    changes.append((binary, hint, description, applier.certainty))

    return changes


def changes_by_description(changes):
    by_description = {}
    for (binary, hint, description, certainty) in changes:
        by_description.setdefault(description, []).append(binary["Package"])
    return by_description


class MultiArchHintFixer(Fixer):
    def __init__(self, hints):
        super(MultiArchHintFixer, self).__init__(name="multiarch-hints")
        self._hints = hints

    def run(
        self,
        basedir,
        package,
        current_version,
        compat_release,
        minimum_certainty=None,
        trust_package=False,
        allow_reformatting=False,
        net_access=True,
        opinionated=False,
        diligence=0,
    ):
        if not net_access:
            # This should never happen - perhaps if something else imported and
            # used this class?
            raise NoChanges(self)
        old_cwd = os.getcwd()
        try:
            os.chdir(basedir)
            changes = apply_multiarch_hints(self._hints, minimum_certainty)
        finally:
            os.chdir(old_cwd)

        overall_certainty = min_certainty(
            [certainty for (binary, hint, description, certainty) in changes]
        )
        by_description = changes_by_description(changes)
        overall_description = ["Apply multi-arch hints."]
        for description, binaries in by_description.items():
            overall_description.append(
                "+ %s: %s" % (", ".join(sorted(binaries)), description)
            )
        return MultiArchFixerResult(
            "\n".join(overall_description), certainty=overall_certainty, changes=changes
        )


APPLIERS = [
    MultiArchHintApplier("ma-foreign", apply_hint_ma_foreign, "certain"),
    MultiArchHintApplier("file-conflict", apply_hint_file_conflict, "certain"),
    MultiArchHintApplier("ma-foreign-library", apply_hint_ma_foreign_lib, "certain"),
    MultiArchHintApplier("dep-any", apply_hint_dep_any, "certain"),
    MultiArchHintApplier("ma-same", apply_hint_ma_same, "certain"),
    MultiArchHintApplier("arch-all", apply_hint_arch_all, "possible"),
]


def report_okay(code, description):
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'description': description}, f)
    logging.info('%s', description)


def report_fatal(code, description):
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'description': description}, f)
    logging.fatal('%s', description)


def main(argv=None):  # noqa: C901
    import argparse
    from breezy.workingtree import WorkingTree

    import breezy  # noqa: E402

    breezy.initialize()
    import breezy.git  # noqa: E402
    import breezy.bzr  # noqa: E402

    from .config import Config

    parser = argparse.ArgumentParser(prog="multi-arch-fixer")
    parser.add_argument(
        "--directory",
        metavar="DIRECTORY",
        help="directory to run in",
        type=str,
        default=".",
    )
    parser.add_argument(
        "--disable-inotify", action="store_true", default=False, help=argparse.SUPPRESS
    )
    parser.add_argument(
        "--identity",
        help="Print user identity that would be used when committing",
        action="store_true",
        default=False,
    )
    # Hide the minimum-certainty option for the moment.
    parser.add_argument(
        "--minimum-certainty",
        type=str,
        choices=SUPPORTED_CERTAINTIES,
        default=None,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + version_string
    )
    parser.add_argument(
        "--allow-reformatting",
        default=None,
        action="store_true",
        help=argparse.SUPPRESS,
    )

    args = parser.parse_args(argv)

    logging.basicConfig(level=logging.INFO, format='%(message)s')

    minimum_certainty = args.minimum_certainty
    wt, subpath = WorkingTree.open_containing(args.directory)
    if args.identity:
        logging.info('%s', get_committer(wt))
        return 0

    update_changelog = args.update_changelog
    allow_reformatting = args.allow_reformatting
    try:
        cfg = Config.from_workingtree(wt, subpath)
    except FileNotFoundError:
        pass
    else:
        if minimum_certainty is None:
            minimum_certainty = cfg.minimum_certainty()
        if allow_reformatting is None:
            allow_reformatting = cfg.allow_reformatting()
        if update_changelog is None:
            update_changelog = cfg.update_changelog()

    use_inotify = ((False if args.disable_inotify else None),)
    try:
        check_clean_tree(wt, wt.basis_tree(), subpath)
    except WorkspaceDirty:
        logging.info("%s: Please commit pending changes first.", wt.basedir)
        return 1

    dirty_tracker = get_dirty_tracker(wt, subpath, use_inotify)
    if dirty_tracker:
        dirty_tracker.mark_clean()

    with cache_download_multiarch_hints() as f:
        hints = multiarch_hints_by_binary(parse_multiarch_hints(f))

    if control_files_in_root(wt, subpath):
        report_fatal(
            "control-files-in-root",
            "control files live in root rather than debian/ " "(LarstIQ mode)",
        )
        return 1

    if is_debcargo_package(wt, subpath):
        report_okay("nothing-to-do", "Package uses debcargo")
        return 0
    if not control_file_present(wt, subpath):
        report_fatal("missing-control-file", "Unable to find debian/control")
        return 1

    try:
        result, summary = run_lintian_fixer(
            wt,
            MultiArchHintFixer(hints),
            update_changelog=update_changelog,
            minimum_certainty=minimum_certainty,
            dirty_tracker=dirty_tracker,
            subpath=subpath,
            allow_reformatting=allow_reformatting,
            net_access=True,
            changes_by="apply-multiarch-hints",
        )
    except NoChanges:
        report_okay("nothing-to-do", "no hints to apply")
        return 0
    except FormattingUnpreservable as e:
        report_fatal(
            "formatting-unpreservable",
            "unable to preserve formatting while editing %s" % e.path,
        )
        return 1
    except GeneratedFile as e:
        report_fatal(
            "generated-file", "unable to edit generated file: %r" % e)
        return 1
    except NotDebianPackage:
        logging.info("%s: Not a debian package.", wt.basedir)
        return 1
    else:
        applied_hints = []
        hint_names = []
        for (binary, hint, description, certainty) in result.changes:
            hint_names.append(hint["link"].split("#")[-1])
            entry = dict(hint.items())
            hint_names.append(entry["link"].split("#")[-1])
            entry["action"] = description
            entry["certainty"] = certainty
            applied_hints.append(entry)
            logging.info("%s: %s" % (binary["Package"], description))
        if os.environ.get('SVP_API') == '1':
            with open(os.environ['SVP_RESULT'], 'w') as f:
                json.dump({
                    'description': "Applied multi-arch hints.",
                    'value': calculate_value(hint_names),
                    'commit-message': 'Apply multi-arch hints',
                    'context': {
                        'applied-hints': applied_hints,
                    }}, f)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python3

from lintian_brush import (
    min_certainty,
    )
from lintian_brush.multiarch_hints import (
    apply_multiarch_hints,
    download_multiarch_hints,
    multiarch_hints_by_binary,
    parse_multiarch_hints,
    )

import os
import sys

if os.environ.get('NET_ACCESS', 'disallow') != 'allow':
    sys.exit(0)

with download_multiarch_hints() as f:
    hints = multiarch_hints_by_binary(parse_multiarch_hints(f))

changes = apply_multiarch_hints(hints, os.environ.get('MINIMUM_CERTAINTY'))

overall_certainty = min_certainty(
    [certainty for (binary, hint, description, certainty) in changes])

print("Apply multi-arch hints.")
print("")

for (binary, hint, description, certainty) in changes:
    print("* %s: %s" % (binary['Package'], description))



#!/usr/bin/python3

# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility for dropping unnecessary constraints."""

import asyncio
import json
import logging
import os
from typing import List, Tuple, Optional, Dict, Callable

from breezy.commit import PointlessCommit, NullCommitReporter
from breezy.workingtree import WorkingTree

from debmutate.control import PkgRelation
from debmutate.deb822 import ChangeConflict
from debmutate.debhelper import MaintscriptEditor
from debmutate.reformatting import FormattingUnpreservable

from debian.changelog import Version
from debian.deb822 import Deb822Dict

from debmutate.control import (
    ControlEditor,
    parse_relations,
    format_relations,
    guess_template_type,
)

from debmutate.reformatting import (
    check_generated_file,
    GeneratedFile,
)

from .changelog import add_changelog_entry
from . import (
    get_committer,
    control_files_in_root,
    control_file_present,
    NotDebianPackage,
    is_debcargo_package,
    )
from .debhelper import drop_obsolete_maintscript_entries


DEFAULT_VALUE_MULTIARCH_HINT = 30


_changelog_policy_noted = False


def _note_changelog_policy(policy, msg):
    global _changelog_policy_noted
    if not _changelog_policy_noted:
        if policy:
            extra = "Specify --no-update-changelog to override."
        else:
            extra = "Specify --update-changelog to override."
        logging.info("%s %s", msg, extra)
    _changelog_policy_noted = True


def depends_obsolete(latest_version: Version, kind: str, req_version: Version):
    req_version = Version(req_version)
    if kind == ">=":
        return latest_version >= req_version
    elif kind == ">>":
        return latest_version > req_version
    elif kind == "=":
        return False
    return False


def conflict_obsolete(latest_version: Version, kind: str, req_version: Version):
    req_version = Version(req_version)
    if kind == "<<":
        return latest_version >= req_version
    elif kind in ("<=", "="):
        return latest_version > req_version
    return False


class UddTimeout(Exception):
    """Timeout while connecting to UDD."""


async def _package_version(package: str, release: str) -> Optional[Version]:
    from .udd import connect_udd_mirror

    try:
        async with await connect_udd_mirror() as conn:
            version = await conn.fetchval(
                "select version from packages where package = $1 and release = $2",
                package, release)
    except asyncio.TimeoutError:
        raise UddTimeout()
    if version is not None:
        return Version(version)
    return None


async def _package_provides(package: str, release: str) -> Optional[List[PkgRelation]]:
    from .udd import connect_udd_mirror

    async with await connect_udd_mirror() as conn:
        provides = await conn.fetchval(
            "select provides from packages where package = $1 and release = $2",
            package, release)
    if provides is not None:
        return [r[1] for r in parse_relations(provides)]
    return None


async def _package_essential(package: str, release: str) -> bool:
    from .udd import connect_udd_mirror

    async with await connect_udd_mirror() as conn:
        return await conn.fetchval(
            "select (essential = 'yes') from packages where package = $1 and release = $2",
            package, release)


async def _package_build_essential(package: str, release: str) -> bool:
    from .udd import connect_udd_mirror

    async with await connect_udd_mirror() as conn:
        depends = await conn.fetchval(
            "select depends from packages where package = $1 and release = $2",
            'build-essential', release)

    build_essential = set()
    for ws1, rel, ws2 in parse_relations(depends):
        build_essential.update([r.name for r in rel])
    return package in build_essential


class PackageChecker(object):

    def __init__(self, release: str, build: bool):
        self.release = release
        self.build = build

    def package_version(self, package: str) -> Optional[Version]:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(_package_version(package, self.release))

    def package_provides(self, package):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(_package_provides(package, self.release))

    def is_essential(self, package: str) -> bool:
        loop = asyncio.get_event_loop()
        if self.build:
            return loop.run_until_complete(_package_build_essential(package, self.release))
        else:
            return loop.run_until_complete(_package_essential(package, self.release))


def drop_obsolete_depends(entry: List[PkgRelation], checker: PackageChecker):
    ors = []
    dropped = []
    for pkgrel in entry:
        newrel = pkgrel
        if pkgrel.version is not None and pkgrel.name != 'debhelper':
            compat_version = checker.package_version(pkgrel.name)
            logging.debug(
                "Relation: %s. Upgrade release %s has %r ",
                pkgrel, checker.release, compat_version)
            if compat_version is not None and depends_obsolete(
                compat_version, *pkgrel.version
            ):
                newrel = PkgRelation.parse(pkgrel.str())[0]
                newrel.version = None
                dropped.append(pkgrel)
                # If the package is essential, we don't need to maintain a
                # dependency on it.
                if checker.is_essential(pkgrel.name):
                    return [], entry
        ors.append(newrel)
    # TODO: if dropped: Check if any ors are implied by existing other
    # dependencies
    return ors, dropped


def drop_obsolete_conflicts(checker: PackageChecker, entry: List[PkgRelation]):
    ors = []
    dropped = []
    for pkgrel in entry:
        if pkgrel.version is not None:
            compat_version = checker.package_version(pkgrel.name)
            if compat_version is not None and conflict_obsolete(
                compat_version, *pkgrel.version
            ):
                dropped.append(pkgrel)
                continue
        ors.append(pkgrel)
    return ors, dropped


def update_depends(base: Deb822Dict, field: str, checker: PackageChecker) -> List[str]:
    return filter_relations(
        base, field,
        lambda oldrelation: drop_obsolete_depends(oldrelation, checker))


def _relations_empty(rels):
    for ws1, rel, ws2 in rels:
        if rel:
            return False
    return True


RelationsCallback = Callable[[List[PkgRelation]], Tuple[List[PkgRelation], List[PkgRelation]]]


def filter_relations(base: Deb822Dict, field: str, cb: RelationsCallback) -> List[str]:
    """Update a relations field."""
    try:
        old_contents = base[field]
    except KeyError:
        return []

    oldrelations = parse_relations(old_contents)
    newrelations = []

    changed = []
    for i, (ws1, oldrelation, ws2) in enumerate(oldrelations):
        relation, dropped = cb(oldrelation)
        changed.extend([d.name for d in dropped])
        if relation == oldrelation or relation:
            newrelations.append((ws1, relation, ws2))
        elif i == 0 and len(oldrelations) > 1:
            # If the first item is removed, then copy the spacing to the next
            # item
            oldrelations[1] = (ws1, oldrelations[1][1], ws2)

    if changed:
        if _relations_empty(newrelations):
            del base[field]
        else:
            base[field] = format_relations(newrelations)
        return changed
    return []


def update_conflicts(base: Deb822Dict, field: str, checker: PackageChecker) -> List[str]:
    return filter_relations(
        base, field,
        lambda oldrelation: drop_obsolete_conflicts(checker, oldrelation))


def drop_old_source_relations(source, compat_release) -> List[Tuple[str, List[str], str]]:
    checker = PackageChecker(compat_release, build=True)
    ret = []
    for field in [
        "Build-Depends",
        "Build-Depends-Indep",
        "Build-Depends-Arch",
    ]:
        packages = update_depends(source, field, checker)
        if packages:
            ret.append((field, packages, compat_release))
    for field in ["Build-Conflicts", "Build-Conflicts-Indep", "Build-Conflicts-Arch"]:
        packages = update_conflicts(source, field, checker)
        if packages:
            ret.append((field, packages, compat_release))
    return ret


def drop_old_binary_relations(binary, upgrade_release: str) -> List[Tuple[str, List[str], str]]:
    checker = PackageChecker(upgrade_release, build=False)
    ret = []
    for field in ["Depends", "Breaks", "Suggests", "Recommends", "Pre-Depends"]:
        packages = update_depends(binary, field, checker)
        if packages:
            ret.append((field, packages, upgrade_release))

    for field in ["Conflicts", "Replaces", "Breaks"]:
        packages = update_conflicts(binary, field, checker)
        if packages:
            ret.append((field, packages, upgrade_release))

    return ret


def drop_old_relations(editor, compat_release: str, upgrade_release: str) -> List[Tuple[Optional[str], List[Tuple[str, List[str], str]]]]:
    dropped: List[Tuple[Optional[str], List[Tuple[str, List[str], str]]]] = []
    source_dropped = []
    try:
        check_generated_file(editor.path)
    except GeneratedFile as e:
        uses_cdbs = (
            e.template_path is not None
            and guess_template_type(e.template_path) == "cdbs"
        )
    else:
        uses_cdbs = False
    if not uses_cdbs:
        source_dropped.extend(drop_old_source_relations(editor.source, compat_release))
    if source_dropped:
        dropped.append((None, source_dropped))

    for binary in editor.binaries:
        binary_dropped = drop_old_binary_relations(binary, upgrade_release)
        if binary_dropped:
            dropped.append((binary["Package"], binary_dropped))

    return dropped


def update_maintscripts(
        wt: WorkingTree, subpath: str, checker: PackageChecker, package: str,
        allow_reformatting: bool = False
        ) -> List[Tuple[str, List[Tuple[int, str, Version]]]]:
    ret = []
    for entry in os.scandir(wt.abspath(os.path.join(subpath))):
        if not (entry.name == "maintscript" or entry.name.endswith(".maintscript")):
            continue
        with MaintscriptEditor(entry.path, allow_reformatting=allow_reformatting) as editor:
            def can_drop(p, v):
                compat_version = checker.package_version(p or package)
                return compat_version is not None and compat_version > v
            removed = drop_obsolete_maintscript_entries(editor, can_drop)
            if removed:
                ret.append((os.path.join(subpath, entry.name), removed))
    return ret


def name_list(packages: List[str]) -> str:
    """Format a list of package names for use in prose.

    Args:
      packages: non-empty list of packages to format
    Returns:
      human-readable string
    """
    if not packages:
        raise ValueError(packages)
    std = list(sorted(set(packages)))
    if len(std) == 1:
        return std[0]
    return ", ".join(std[:-1]) + " and " + std[-1]


class ScrubObsoleteResult(object):

    def __init__(self, specific_files, maintscript_removed, control_removed):
        self.specific_files = specific_files
        self.maintscript_removed = maintscript_removed
        self.control_removed = control_removed

    def __bool__(self):
        return bool(self.control_removed) or bool(self.maintscript_removed)

    def value(self) -> int:
        value = DEFAULT_VALUE_MULTIARCH_HINT
        for para, changes in self.control_removed:
            for field, packages, release in changes:
                value += len(packages) * 2
        for path, removed, release in self.maintscript_removed:
            value += len(removed)
        return value

    def itemized(self) -> Dict[str, List[str]]:
        summary: Dict[str, List[str]] = {}
        for para, changes in self.control_removed:
            for field, packages, release in changes:
                if para:
                    summary.setdefault(release, []).append(
                        "%s: Drop versioned constraint on %s in %s."
                        % (para, name_list(packages), field)
                    )
                else:
                    summary.setdefault(release, []).append(
                        "%s: Drop versioned constraint on %s."
                        % (field, name_list(packages))
                    )
        if self.maintscript_removed:
            total_entries = sum(
                [len(entries) for name, entries, release in self.maintscript_removed])
            summary.setdefault(self.maintscript_removed[0][2], []).append(
                "Remove %d maintscript entries from %d files." % (
                    total_entries, len(self.maintscript_removed))
            )
        return summary


def _scrub_obsolete(
        wt: WorkingTree,
        debian_path: str,
        compat_release: str,
        upgrade_release: str,
        allow_reformatting: bool = True) -> ScrubObsoleteResult:
    specific_files = []
    control_path = os.path.join(debian_path, "control")
    try:
        with ControlEditor(
                wt.abspath(control_path),
                allow_reformatting=allow_reformatting) as editor:
            specific_files.append(control_path)
            package = editor.source["Source"]
            control_removed = drop_old_relations(editor, compat_release, upgrade_release)
    except FileNotFoundError:
        if wt.has_filename(os.path.join(debian_path, "debcargo.toml")):
            control_removed = []
        else:
            raise NotDebianPackage(wt, debian_path)

    maintscript_removed = []
    for path, removed in update_maintscripts(
            wt, debian_path, PackageChecker(upgrade_release, build=False), package,
            allow_reformatting):
        if removed:
            maintscript_removed.append((path, removed, upgrade_release))
            specific_files.append(path)

    return ScrubObsoleteResult(
        specific_files=specific_files,
        control_removed=control_removed,
        maintscript_removed=maintscript_removed,
    )


def scrub_obsolete(
        wt: WorkingTree, subpath: str, compat_release: str, upgrade_release: str,
        update_changelog=None,
        allow_reformatting: bool = False) -> ScrubObsoleteResult:
    """Scrub obsolete entries.
    """

    if control_files_in_root(wt, subpath):
        debian_path = subpath
    else:
        debian_path = os.path.join(subpath, 'debian')

    result = _scrub_obsolete(
        wt, debian_path, compat_release, upgrade_release, allow_reformatting)

    if not result:
        return result

    specific_files = list(result.specific_files)
    summary = result.itemized()

    changelog_path = os.path.join(debian_path, "changelog")

    if update_changelog is None:
        from .detect_gbp_dch import guess_update_changelog
        from debian.changelog import Changelog

        with wt.get_file(changelog_path) as f:
            cl = Changelog(f, max_blocks=1)

        dch_guess = guess_update_changelog(wt, debian_path, cl)
        if dch_guess:
            update_changelog = dch_guess[0]
            _note_changelog_policy(update_changelog, dch_guess[1])
        else:
            # Assume we should update changelog
            update_changelog = True

    if update_changelog:
        lines = []
        for release, entries in summary.items():
            lines.append("Remove constraints unnecessary since %s:" % release)
            lines.extend(["+ " + line for line in entries])
        add_changelog_entry(wt, changelog_path, lines)
        specific_files.append(changelog_path)

    lines = []
    for release, entries in summary.items():
        lines.extend(["Remove constraints unnecessary since %s" % release, ""])
        lines.extend(["* " + line for line in entries])
    lines.extend(["", "Changes-By: deb-scrub-obsolete"])

    committer = get_committer(wt)

    try:
        wt.commit(
            specific_files=specific_files,
            message="\n".join(lines),
            allow_pointless=False,
            reporter=NullCommitReporter(),
            committer=committer,
        )
    except PointlessCommit:
        pass

    return result


def report_fatal(code: str, description: str) -> None:
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'description': description}, f)
    logging.fatal('%s', description)


def report_okay(code: str, description: str) -> None:
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'description': description}, f)
    logging.info('%s', description)


def main():  # noqa: C901
    import argparse
    import breezy  # noqa: E402

    breezy.initialize()
    import breezy.git  # noqa: E402
    import breezy.bzr  # noqa: E402

    from breezy.workspace import (
        check_clean_tree,
        WorkspaceDirty,
        )
    from . import (
        version_string,
    )
    from .config import Config

    parser = argparse.ArgumentParser(prog="deb-scrub-obsolete")
    parser.add_argument(
        "--directory",
        metavar="DIRECTORY",
        help="directory to run in",
        type=str,
        default=".",
    )
    parser.add_argument(
        "--upgrade-release",
        metavar="UPGRADE-RELEASE",
        help="Release to allow upgrading from.",
        default="oldstable",
    )
    parser.add_argument(
        '--compat-release',
        metavar='COMPAT-RELEASE',
        help="Release to allow building on.",
        default=None)
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )
    parser.add_argument(
        "--allow-reformatting",
        default=None,
        action="store_true",
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + version_string
    )
    parser.add_argument(
        "--identity",
        help="Print user identity that would be used when committing",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug", help="Describe all considerd changes.", action="store_true"
    )

    args = parser.parse_args()

    wt, subpath = WorkingTree.open_containing(args.directory)
    if args.identity:
        logging.info('%s', get_committer(wt))
        return 0

    try:
        check_clean_tree(wt, wt.basis_tree(), subpath)
    except WorkspaceDirty:
        logging.info("%s: Please commit pending changes first.", wt.basedir)
        return 1

    import distro_info
    debian_info = distro_info.DebianDistroInfo()
    upgrade_release = debian_info.codename(args.upgrade_release)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO, format='%(message)s')

    update_changelog = args.update_changelog
    allow_reformatting = args.allow_reformatting
    if args.compat_release:
        compat_release = debian_info.codename(args.compat_release)
    else:
        compat_release = None
    try:
        cfg = Config.from_workingtree(wt, subpath)
    except FileNotFoundError:
        pass
    else:
        if update_changelog is None:
            update_changelog = cfg.update_changelog()
        if allow_reformatting is None:
            allow_reformatting = cfg.allow_reformatting()
        if compat_release is None:
            compat_release = cfg.compat_release()

    if compat_release is None:
        compat_release = debian_info.codename('oldstable')

    if upgrade_release != compat_release:
        logging.info(
            "Removing run time constraints unnecessary since %s"
            " and build time constraints unnecessary since %s",
            upgrade_release, compat_release)
    else:
        logging.info(
            "Removing run time and build time constraints unnecessary "
            "since %s", compat_release)

    if allow_reformatting is None:
        allow_reformatting = False

    if is_debcargo_package(wt, subpath):
        report_fatal("nothing-to-do", "Package uses debcargo")
        return 1
    elif not control_file_present(wt, subpath):
        report_fatal("missing-control-file", "Unable to find debian/control")
        return 1

    try:
        result = scrub_obsolete(
            wt, subpath, compat_release, upgrade_release,
            update_changelog=args.update_changelog,
            allow_reformatting=allow_reformatting
        )
    except FormattingUnpreservable as e:
        report_fatal(
            "formatting-unpreservable",
            "unable to preserve formatting while editing %s" % e.path,
        )
        return 1
    except GeneratedFile as e:
        report_fatal(
            "generated-file", "unable to edit generated file: %r" % e
        )
        return 1
    except NotDebianPackage:
        report_fatal('not-debian-package', 'Not a Debian package.')
        return 1
    except ChangeConflict as e:
        report_fatal('change-conflict', 'Generated file changes conflict: %s' % e)
        return 1
    except UddTimeout:
        report_fatal('udd-timeout', 'Timeout communicating with UDD')
        return 1

    if not result:
        report_okay("nothing-to-do", "no obsolete constraints")
        return 0

    if os.environ.get("SVP_API") == "1":
        with open(os.environ["SVP_RESULT"], "w") as f:
            json.dump({
                "description": "Remove constraints unnecessary since %s."
                % upgrade_release,
                "value": result.value(),
                "context": {
                    "specific_files": result.specific_files,
                    "maintscript_removed": result.maintscript_removed,
                    "control_removed": result.control_removed,
                }
            }, f)

    logging.info("Scrub obsolete settings.")
    for release, lines in result.itemized().items():
        for line in lines:
            logging.info("* %s", line)

    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main())



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Lintian-brush configuration file."""

import os

from configobj import ConfigObj
import warnings
import distro_info


PACKAGE_CONFIG_FILENAME = "debian/lintian-brush.conf"


SUPPORTED_KEYS = [
    "compat-release",
    "minimum-certainty",
    "allow-reformatting",
    "update-changelog",
]


def resolve_release_codename(name):
    debian = distro_info.DebianDistroInfo()
    if debian.codename(name):
        return debian.codename(name)
    if debian.valid(name):
        return name
    ubuntu = distro_info.UbuntuDistroInfo()
    if ubuntu.valid(name):
        return name
    return None


class Config(object):
    """A configuration file."""

    def __init__(self, path):
        if not os.path.exists(path):
            raise FileNotFoundError(path)
        self._obj = ConfigObj(path, raise_errors=True, file_error=True)
        for k in self._obj.keys():
            if k not in SUPPORTED_KEYS:
                warnings.warn("unknown setting %s in %s" % (k, path))

    @classmethod
    def from_workingtree(cls, wt, subpath):
        return cls(os.path.join(wt.basedir, subpath, PACKAGE_CONFIG_FILENAME))

    def compat_release(self):
        value = self._obj.get("compat-release")
        codename = resolve_release_codename(value)
        if codename is None:
            warnings.warn("unknown compat release %s, ignoring." % value)
        return codename

    def allow_reformatting(self):
        try:
            return self._obj.as_bool("allow-reformatting")
        except KeyError:
            return None

    def minimum_certainty(self):
        return self._obj.get("minimum-certainty")

    def update_changelog(self):
        try:
            return self._obj.as_bool("update-changelog")
        except KeyError:
            return None



#!/usr/bin/python3

from debmutate.reformatting import check_generated_file
from lintian_brush.fixer import report_result, LintianIssue
from lintian_brush.line_editor import LineEditor

check_generated_file('debian/rules')

try:
    with LineEditor('debian/rules', 'b') as e:
        for lineno, line in e:
            newline = line.replace(b'$(PWD)', b'$(CURDIR)')
            if newline != line:
                issue = LintianIssue(
                    'source',
                    'debian-rules-calls-pwd', info='line %d' % lineno)
                if issue.should_fix():
                    e[lineno] = newline
                    issue.report_fixed()
except FileNotFoundError:
    pass

report_result("debian/rules: Avoid using $(PWD) variable.")



#!/usr/bin/python3

import re
from lintian_brush.fixer import report_result, opinionated, fixed_lintian_tag
from debmutate._rules import update_rules, Makefile
from typing import Set, Tuple, Optional

architecture_included = False
PATH = '/usr/share/dpkg/architecture.mk'
INCLUDE_LINE = b'include ' + PATH.encode('ascii')

_variables: Set[str] = set()
message = None


def variable_defined(var):
    global _variables
    if _variables:
        return var in _variables
    k = {}
    with open(PATH, 'r') as f:
        for line in f:
            if line.strip().startswith('$(foreach '):
                vs = line.strip()[len('$(foreach '):].split(',')
                k[vs[0]] = vs[1]
    for machine in k['machine'].split(' '):
        for var in k['var'].split(' '):
            _variables.add('DEB_%s_%s' % (machine, var))


def is_dpkg_architecture_line(line: str) -> Tuple[
        Optional[str], bool, bool, Optional[bool]]:
    m = re.match(b'([A-Z_]+)[ \t]*([:?]?=)[ \t]*(.*)', line.strip())
    if not m:
        return (None, False, False, None)
    variable = m.group(1).decode()
    hard_assignment = m.group(2) != b"?="
    assignment = m.group(3)
    if not variable_defined(variable):
        return (variable, False, False, hard_assignment)
    m = re.match(b'\\$\\(shell dpkg-architecture -q([A-Z_]+)\\)', assignment)
    if not m or variable != m.group(1).decode():
        return (variable, True, False, hard_assignment)
    else:
        return (variable, True, True, hard_assignment)


def drop_arch_line(line):
    global architecture_included
    if line.strip() == INCLUDE_LINE:
        architecture_included = True
        return line
    (variable, variable_matches,
     value_matches, hard) = is_dpkg_architecture_line(line)
    if not variable_matches:
        return line
    if not value_matches:
        # The assignment is different; let's include
        # the mk file *before* this assignment
        if not architecture_included and variable_matches:
            architecture_included = True
            return [INCLUDE_LINE, line]
        else:
            return line
    if hard:
        lineno = -1  # TODO(jelmer): Pass this up
        fixed_lintian_tag(
            'source', 'debian-rules-sets-dpkg-architecture-variable',
            info='%s (line %d)' % (variable, lineno))
    if not architecture_included:
        architecture_included = True
        return INCLUDE_LINE
    return None


def update_assignment_kind(line):
    global architecture_included, message
    if line.strip() == INCLUDE_LINE:
        architecture_included = True
        return line
    (variable, variable_matches,
     value_matches, hard) = is_dpkg_architecture_line(line)
    if not variable_matches:
        return line
    if not value_matches:
        return line
    if not hard:
        if architecture_included:
            return None
        else:
            # Nothing wrong here
            return line
    lineno = -1  # TODO(jelmer): Pass this up
    fixed_lintian_tag(
        'source', 'debian-rules-sets-dpkg-architecture-variable',
        info='%s (line %d)' % (variable, lineno))
    if architecture_included:
        message = 'Rely on existing architecture.mk include.'
        return None
    return re.sub(b'([:?]?=)', b'?=', line)


if opinionated():
    try:
        mf = Makefile.from_path('debian/rules')
    except FileNotFoundError:
        pass
    else:
        for line in mf.contents:
            if (isinstance(line, bytes) and
                    is_dpkg_architecture_line(line)[1:3] == (True, True)):
                update_rules(global_line_cb=drop_arch_line)
                message = (
                    'Rely on pre-initialized dpkg-architecture variables.')
                break
else:
    message = 'Use ?= for assignments to architecture variables.'
    update_rules(global_line_cb=update_assignment_kind)

report_result(message)



#!/usr/bin/python3


from debmutate.copyright import CopyrightEditor, NotMachineReadableError
from lintian_brush.fixer import report_result, LintianIssue


def swap_files_glob(copyright):
    files_i = 0
    for i, paragraph in enumerate(copyright._Copyright__paragraphs):
        if "Files" in paragraph:
            if paragraph["Files"] == "*" and files_i > 0:
                issue = LintianIssue(
                    'source',
                    'global-files-wildcard-not-first-paragraph-in-'
                    'dep5-copyright')
                if issue.should_fix():
                    copyright._Copyright__paragraphs.insert(
                        0, copyright._Copyright__paragraphs.pop(i))
                    issue.report_fixed()
            files_i += 1


try:
    with CopyrightEditor() as updater:
        swap_files_glob(updater.copyright)
except (FileNotFoundError, NotMachineReadableError):
    pass

report_result('Make "Files: *" paragraph the first in the copyright file.')



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Debhelper utility functions."""

try:
    from functools import cache
except ImportError:
    # Python < 3.8
    from functools import lru_cache

    def cache(user_function):
        return lru_cache(maxsize=None)(user_function)

import json
import os
import subprocess
from typing import Dict, Optional, List, Callable, Tuple

from debian.changelog import Version
from debmutate.control import drop_dependency, add_dependency
from debmutate._rules import update_rules, dh_invoke_drop_with, dh_invoke_add_with

DEBHELPER_BUILD_STEPS = ["configure", "build", "test", "install", "clean"]


def detect_debhelper_buildsystem(step: Optional[str] = None) -> Optional[str]:
    """Detect the build system for debhelper

    Args:
      step: Optional step to determine the buildsystem for
    Returns:
      Build system name or None, if none could be found
    """
    if os.path.exists("configure.ac") or os.path.exists("configure.in"):
        return "autoconf"
    env = dict(os.environ)
    # Necessary for debhelper (<= 13.5.2), or it'll write debian/.debhelper
    # files.
    env['DH_NO_ACT'] = '1'
    output = subprocess.check_output(
        ["dh_assistant", "which-build-system"], env=env).decode()
    return json.loads(output)["build-system"]


@cache
def _get_lintian_compat_levels() -> Dict[str, int]:
    output = subprocess.check_output(
        ["dh_assistant", "supported-compat-levels"]).decode()
    return json.loads(output)


def lowest_non_deprecated_compat_level() -> int:
    """Find the lowest non-deprecated debhelper compat level."""
    return _get_lintian_compat_levels()["LOWEST_NON_DEPRECATED_COMPAT_LEVEL"]


def highest_stable_compat_level() -> int:
    """Find the highest stable debhelper compat level."""
    return _get_lintian_compat_levels()["HIGHEST_STABLE_COMPAT_LEVEL"]


def maximum_debhelper_compat_version(compat_release: str) -> int:
    """Retrieve the maximum supported debhelper compat version fior a release.

    Args:
      compat_release: A release name (Debian or Ubuntu, currently)
    Returns:
      debhelper compat version
    """
    from .release_info import key_package_version

    debhelper_version = key_package_version("debhelper", compat_release)
    if debhelper_version is None:
        max_version = lowest_non_deprecated_compat_level()
    else:
        max_version = int(str(debhelper_version).split(".")[0])
    return max_version


def write_rules_template(
        path: str, buildsystem: Optional[str] = None,
        addons: Optional[List[str]] = None,
        env: Optional[Dict[str, str]] = None) -> None:
    if addons is None:
        addons = []
    dh_args = ["$@"]
    if buildsystem:
        dh_args.append("--buildsystem=%s" % buildsystem)
    for addon in addons:
        dh_args.append("--with=%s" % addon)

    with open(path, "w") as f:
        f.write(
            """\
#!/usr/bin/make -f

"""
        )
        if env:
            for key, value in env.items():
                f.write("export %s := %s\n" % (key, value))
            f.write("\n")

        f.write(
            """\
%:
\tdh """
            + " ".join(dh_args)
            + """
"""
        )
    os.chmod(path, 0o755)


def drop_obsolete_maintscript_entries(
        editor, should_remove: Callable[[str, Version], bool]
        ) -> List[Tuple[int, str, Version]]:
    """Drop obsolete entries from a maintscript file.

    Args:
      editor: editor to use to access the maintscript
      should_remove: callable to check whether a package/version tuple is obsolete
    Returns:
      list of tuples with index, package, version of entries that were removed
    """
    remove: List[Tuple[int, str, Version]] = []
    comments = []
    entries_removed = []
    for i, entry in enumerate(list(editor.lines)):
        if isinstance(entry, str):
            comments.append((i, None, None))
            continue
        prior_version = getattr(entry, "prior_version", None)
        if prior_version is not None:
            if should_remove(entry.package, Version(prior_version)):
                remove.extend(comments)
                remove.append((i, entry.package, Version(prior_version)))
                entries_removed.append(i)
        comments = []
    removed = []
    for i, pkg, version in reversed(remove):
        removed.append(editor.lines[i])
        del editor.lines[i]
    return entries_removed


def drop_sequence(control, rules, sequence):
    new_depends = drop_dependency(
        control.source.get("Build-Depends", ""), "dh-" + sequence)
    if new_depends != control.source['Build-Depends']:
        def drop_with(line, target):
            return dh_invoke_drop_with(line, sequence.replace('-', '_').encode())
        update_rules(drop_with)
    new_depends = drop_dependency(
        new_depends, "dh-sequence-" + sequence)
    if control['Build-Depends'] == new_depends:
        return False
    control['Build-Depends'] = new_depends
    return True


def add_sequence(control, rules, sequence):
    control.source['Build-Depends'] = add_dependency(
        control.source.get('Build-Depends'), 'dh-vim-addon')

    def add_with(line, target):
        if line.startswith(b'dh ') or line.startswith(b'dh_'):
            return dh_invoke_add_with(line, sequence.replace('-', '_').encode())
        return line
    update_rules(add_with)



#!/usr/bin/python3

from debmutate._rules import RulesEditor
from lintian_brush.fixer import report_result, LintianIssue

certainty = None

try:
    with RulesEditor() as updater:
        for rule in updater.makefile.iter_rules(b'get-orig-source'):
            commands = rule.commands()
            if [b'uscan'] == [c.split(b' ')[0] for c in commands]:
                certainty = 'certain'
            else:
                certainty = 'possible'
            issue = LintianIssue(
                'source',
                'debian-rules-contains-unnecessary-get-orig-source-target',
                '')
            if issue.should_fix():
                rule.clear()
                updater.makefile.drop_phony(b'get-orig-source')
                issue.report_fixed()
except FileNotFoundError:
    pass


report_result(
    'Remove unnecessary get-orig-source-target.',
    certainty=certainty)



#!/usr/bin/python3

from lintian_brush.systemd import SystemdServiceEditor, systemd_service_files
from lintian_brush.fixer import report_result, fixed_lintian_tag


for path in systemd_service_files():
    with SystemdServiceEditor(path) as updater:
        if (updater.file['Unit']['DefaultDependencies'] == 'no' and
                'shutdown.target' in updater.file['Unit']['Conflicts'] and
                'shutdown.target' not in updater.file['Unit']['Before']):
            updater.file['Unit']['Before'].append('shutdown.target')
            fixed_lintian_tag(
                'source', 'systemd-service-file-shutdown-problems', path)

report_result("Add Before=shutdown.target to Unit section.")



#!/usr/bin/python3

from functools import partial
import os
import re
from typing import Dict

from debian.copyright import License, NotMachineReadableError

from debmutate.copyright import CopyrightEditor
from lintian_brush.fixer import report_result, fixed_lintian_tag

SYNOPSIS_ALIAS: Dict[str, str] = {}
updated = set()


def replace_symlink_path(synopsis, m):
    path = m.group(0)
    waslink = os.path.islink(path)
    synopsis = SYNOPSIS_ALIAS.get(synopsis, synopsis)
    newpath = '/usr/share/common-licenses/' + synopsis.rstrip('+')
    if not os.path.exists(newpath) or os.path.islink(newpath):
        return path
    if not newpath.startswith(path + '-'):
        return path
    updated.add(synopsis)
    if waslink:
        fixed_lintian_tag(
            'all', 'copyright-refers-to-symlink-license',
            info=path.lstrip('/'))
    fixed_lintian_tag(
        'all', 'copyright-refers-to-versionless-license-file',
        info=path.lstrip('/'))
    return newpath


try:
    with CopyrightEditor() as updater:
        for para in updater.copyright.all_paragraphs():
            license = para.license
            if not license or not license.text:
                continue
            changed_text = re.sub(
                '/usr/share/common-licenses/([A-Za-z0-9-.]+)',
                partial(replace_symlink_path, license.synopsis), license.text)
            if changed_text != license.text:
                para.license = License(license.synopsis, changed_text)
except (FileNotFoundError, NotMachineReadableError):
    pass


report_result(
    'Refer to specific version of license %s.' % ', '.join(sorted(updated)))



#!/usr/bin/python3

import os
import sys
from lintian_brush.fixer import (
    package_is_native,
    report_result,
    LintianIssue,
    opinionated,
    )

if not package_is_native():
    # Nothing to do
    sys.exit(0)

if not opinionated():
    sys.exit(0)

for name in ['debian/upstream/signing-key.asc']:
    if not os.path.exists(name):
        continue
    issue = LintianIssue(
        'source', 'public-upstream-key-in-native-package', ())
    if not issue.should_fix():
        continue
    issue.report_fixed()
    os.unlink(name)

if os.listdir('debian/upstream') == []:
    os.rmdir('debian/upstream')

report_result(
    'Remove upstream signing key in native source package.',
    certainty='certain')



#!/usr/bin/python3

from email.utils import parseaddr
from lintian_brush.fixer import control, report_result, fixed_lintian_tag

QA_MAINTAINER = "Debian QA Group <packages@qa.debian.org>"

with control as updater:
    try:
        email = parseaddr(updater.source["Maintainer"])[1]
    except KeyError:
        # No maintainer? Weird, but sure.
        pass
    else:
        if email == "packages@qa.debian.org":
            updater.source["Maintainer"] = QA_MAINTAINER
            fixed_lintian_tag(
                updater.source, 'wrong-debian-qa-group-name')

report_result("Fix Debian QA group name.")



#!/usr/bin/python3

import sys

from debmutate.control import parse_standards_version
from lintian_brush.fixer import control, report_result, LintianIssue
from lintian_brush.standards_version import iter_standards_versions

try:
    release_dates = dict(iter_standards_versions())
except FileNotFoundError:
    sys.exit(2)


with control as updater:
    try:
        sv = parse_standards_version(updater.source['Standards-Version'])
    except KeyError:
        sys.exit(0)
    if (sv in release_dates or
            sv[:4] in release_dates or
            len(sv) == 3 and
            sv + (0, ) in release_dates):
        sys.exit(0)
    invalid_version = updater.source['Standards-Version']
    issue = LintianIssue(
        'source', 'invalid-standards-version', invalid_version)
    if issue.should_fix():
        issue.report_fixed()
        if len(sv) == 2 and (sv[0], sv[1], 0, 0) in release_dates:
            updater.source['Standards-Version'] += '.0'
            report_result("Add missing .0 suffix in Standards-Version.")
        elif sv > sorted(release_dates)[-1]:
            # Maybe we're just unaware of new policy releases?
            sys.exit(0)
        else:
            # Just find the previous standards version..
            candidates = [v for v in release_dates if v < sv]
            newsv = sorted(candidates)[-1]
            newsv_str = '.'.join([str(x) for x in newsv])
            report_result(
                'Replace invalid standards version %s with valid %s.' % (
                    updater.source['Standards-Version'], newsv_str))
            updater.source['Standards-Version'] = newsv_str



#!/usr/bin/python3

import re
import sys

from lintian_brush.fixer import LintianIssue, report_result


def expand_tabs(line, tabwidth=8):
    ret = []
    for i, c in enumerate(line):
        if line[i:i+1] == b'\t':
            # Round up to the next unit of tabwidth
            ret.extend([b' '] * (tabwidth - len(ret) % tabwidth))
        else:
            ret.append(line[i:i+1])
    return b''.join(ret)


def whitespace_prefix_length(line):
    return len(re.match(b'^\\s*', line).group(0))


def value_offset(line):
    if not line.strip():
        return None
    if line.startswith(b'#'):
        return None
    if line[0:1] in (b'\t', b' '):
        return whitespace_prefix_length(line)
    else:
        try:
            key, value = line.split(b':', 1)
        except ValueError:
            return None
        else:
            return len(key) + 1 + whitespace_prefix_length(value)


EXPECTED_HEADER = (
    b'Format: '
    b'https://www.debian.org/doc/packaging-manuals/copyright-format/1.0')


UNICODE_LINE_BREAK = "\u2028".encode('utf-8')
UNICODE_PARAGRAPH_SEPARATOR = "\u2029".encode('utf-8')


tabs_replaced = False
unicode_linebreaks_replaced = False

try:
    lines = []
    with open('debian/copyright', 'rb') as f:
        line = f.readline()
        if line.rstrip().rstrip(b'/') != EXPECTED_HEADER:
            # Not a machine-readable copyright file
            sys.exit(0)
        lines.append(line)
        prev_value_offset = None
        for lineno, line in enumerate(f, start=2):
            if line.startswith(b'\t'):
                issue = LintianIssue(
                    'source', 'tab-in-license-text',
                    info='debian/copyright (paragraph at line %d)' % lineno)
                if issue.should_fix():
                    for option in [
                            b' \t' + line[1:], b' \t' + line[2:],
                            b' ' * 8 + line[1:]]:
                        if value_offset(option) == prev_value_offset:
                            line = option
                            break
                    else:
                        line = b' \t' + line[1:]
                    tabs_replaced = True
                    issue.report_fixed()
            if UNICODE_PARAGRAPH_SEPARATOR in line:
                # Not quite the same thing, but close enough..
                line = line.replace(
                    UNICODE_PARAGRAPH_SEPARATOR,
                    UNICODE_LINE_BREAK * 2)
            if UNICODE_LINE_BREAK in line:
                unicode_linebreaks_replaced = True
                parts = line.split(UNICODE_LINE_BREAK)
                # If the line is empty, replace it with a .
                parts = [p if p else b'.' for p in parts]
                parts = [parts[0]] + [b' ' + p for p in parts[1:]]
                line = b'\n'.join(parts)
            lines.append(line)
            prev_value_offset = value_offset(line)
except FileNotFoundError:
    pass
else:
    with open('debian/copyright', 'wb') as f:
        f.writelines(lines)

sys.stdout.write('debian/copyright: ')
if tabs_replaced:
    sys.stdout.write('use spaces rather than tabs to start continuation lines')
    if unicode_linebreaks_replaced:
        sys.stdout.write(', ')
if unicode_linebreaks_replaced:
    sys.stdout.write('replace unicode linebreaks with regular linebreaks')
sys.stdout.write('.\n')
report_result()



#!/usr/bin/python3

from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    report_result,
    )

import sys

# TODO(jelmer): Support unknown-priority tag

certainty = 'certain'
updated = {}

with control as updater:
    if updater.source.get('Priority'):
        sys.exit(0)
    binary_priorities = set()
    for binary in updater.binaries:
        if not binary.get('Priority'):
            # TODO(jelmer): Check if all dependencies are priority standard or
            # important and if they are, consider bump the priority or not
            # setting the priority at all?
            binary['Priority'] = 'optional'
            certainty = 'confident'
            updated[binary['Package']] = binary['Priority']
            fixed_lintian_tag(binary, 'recommended-field', 'Priority')
        if binary.get('Priority'):
            binary_priorities.add(binary['Priority'])
    if len(binary_priorities) == 1:
        updater.source['Priority'] = binary_priorities.pop()
        for binary in updater.binaries:
            try:
                del binary['Priority']
            except KeyError:
                pass
        report_result(
            'Set priority in source stanza, since it is the same '
            'for all packages.',
            certainty='confident')
    elif updated:
        report_result(
            'Set priority for binary packages %s.' % (
                ['%s (%s)' % item for item in updated.items()]))



#!/usr/bin/python3

import os
import sys
from lintian_brush.fixer import LintianIssue, report_result

OLD_PATH = 'debian/tests/control.autodep8'
NEW_PATH = 'debian/tests/control'

if not os.path.exists(OLD_PATH):
    sys.exit(0)


issue = LintianIssue(
    'source', 'debian-tests-control-autodep8-is-obsolete',
    info=OLD_PATH)

if not issue.should_fix():
    sys.exit(0)

if not os.path.exists(NEW_PATH):
    os.rename(OLD_PATH, 'debian/tests/control')
    issue.report_fixed()
    report_result("Rename obsolete path %s to %s." % (OLD_PATH, NEW_PATH))
else:
    merge_issue = LintianIssue(
        'source', 'debian-tests-control-and-control-autodep8',
        info='%s %s' % (OLD_PATH, NEW_PATH))
    if not merge_issue.should_fix():
        sys.exit(0)
    merge_issue.report_fixed()
    issue.report_fixed()
    with open(NEW_PATH, 'ab') as outf:
        outf.write(b'\n')
        with open(OLD_PATH, 'rb') as inf:
            outf.writelines(inf.readlines())
    os.unlink(OLD_PATH)
    report_result("Merge %s into %s." % (OLD_PATH, NEW_PATH))



#!/usr/bin/python3

from debmutate.control import (
    get_relation,
    add_dependency,
    drop_dependency,
    )
from lintian_brush.fixer import control, report_result


with control as updater:
    try:
        pos, old = get_relation(
            updater.source.get('Build-Depends-Indep', ''), 'debhelper-compat')
    except KeyError:
        pass
    else:
        updater.source['Build-Depends'] = add_dependency(
            updater.source.get('Build-Depends', ''), old)
        updater.source['Build-Depends-Indep'] = drop_dependency(
            updater.source.get('Build-Depends-Indep', ''), 'debhelper-compat')
        if not updater.source['Build-Depends-Indep'].strip():
            del updater.source['Build-Depends-Indep']


report_result(
    'Move debhelper-compat from Build-Depends-Indep to Build-Depends.')



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.config."""

from breezy.tests import (
    TestCase,
    TestCaseWithTransport,
)

from distro_info import DebianDistroInfo

from lintian_brush.config import (
    Config,
    resolve_release_codename,
)


class ResolveCodeNameTests(TestCase):
    def test_resolve_debian(self):
        self.assertEqual("sid", resolve_release_codename("sid"))
        self.assertEqual("buster", resolve_release_codename("buster"))
        self.assertEqual("sid", resolve_release_codename("unstable"))

    def test_resolve_unknown(self):
        self.assertEqual(None, resolve_release_codename("blah"))

    def test_resolve_ubuntu(self):
        self.assertEqual("trusty", resolve_release_codename("trusty"))


class ConfigReadTests(TestCaseWithTransport):
    def test_compat_release(self):
        debian = DebianDistroInfo()
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/lintian-brush.conf",
                    """\
compat-release = testing
""",
                ),
            ]
        )
        cfg = Config("debian/lintian-brush.conf")
        self.assertEqual(debian.testing(), cfg.compat_release())

    def test_minimum_certainty(self):
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/lintian-brush.conf",
                    """\
minimum-certainty = possible
""",
                ),
            ]
        )
        cfg = Config("debian/lintian-brush.conf")
        self.assertEqual("possible", cfg.minimum_certainty())

    def test_update_changelog(self):
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/lintian-brush.conf",
                    """\
update-changelog = True
""",
                ),
            ]
        )
        cfg = Config("debian/lintian-brush.conf")
        self.assertEqual(True, cfg.update_changelog())

    def test_unknown(self):
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/lintian-brush.conf",
                    """\
unknown = dunno
""",
                ),
            ]
        )
        with self.assertWarns(Warning):
            Config("debian/lintian-brush.conf")

    def test_missing(self):
        self.assertRaises(FileNotFoundError, Config, "blah.conf")



#!/usr/bin/python3

from lintian_brush.fixer import report_result, fixed_lintian_tag
from lintian_brush.systemd import systemd_service_files, SystemdServiceEditor
import os

for path in systemd_service_files():
    base, required_ext = os.path.splitext(path)
    with SystemdServiceEditor(path) as updater:
        for i, alias in enumerate(updater.file['Unit']['Alias']):
            base, ext = os.path.splitext(alias)
            if ext != required_ext:
                updater.file['Unit']['Alias'][i] = base + required_ext
                fixed_lintian_tag(
                    'source',
                    'systemd-service-alias-without-extension',
                    '%s' % path)


report_result('Use proper extensions in Alias in systemd files.')



#!/usr/bin/python3

import email.utils
from debmutate.changelog import ChangelogEditor

from lintian_brush.fixer import report_result, LintianIssue, warn

versions = []


with ChangelogEditor() as updater:
    for block in updater.changelog:
        try:
            dt = email.utils.parsedate_to_datetime(block.date)
        except (TypeError, ValueError):
            warn('Invalid date %r for %s' % (block.date, block.version))
            # parsedate_to_datetime is buggy and raises a TypeError
            # when the date is invalid.
            continue
        if dt is None:
            # Can't interpret the date. Just ignore..
            continue
        newdate = email.utils.format_datetime(dt)
        issue = LintianIssue(
            'source', 'debian-changelog-has-wrong-day-of-week',
            info='%04d-%02d-%02d is a %s' % (
                dt.year, dt.month, dt.day, dt.strftime('%A')))
        if newdate[:3] != block.date[:3] and issue.should_fix():
            block.date = newdate
            versions.append(block.version)
            issue.report_fixed()

if len(versions) == 1:
    report_result(
        'Fix day-of-week for changelog entry %s.'
        % ', '.join([str(v) for v in versions]))
else:
    report_result(
        'Fix day-of-week for changelog entries %s.'
        % ', '.join([str(v) for v in versions]))



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import control, report_result, LintianIssue
from debmutate._rules import RulesEditor


def get_archs():
    archs = set()

    # TODO(jelmer): Not really an 'update'
    with control as updater:
        for binary in updater.binaries:
            archs.add(binary['Architecture'])
    return archs


added = []


if not os.path.exists('debian/rules'):
    sys.exit(2)
with RulesEditor() as editor:
    has_build_arch = bool(list(editor.makefile.iter_rules(b'build-arch', exact=False)))
    has_build_indep = bool(list(editor.makefile.iter_rules(b'build-indep', exact=False)))

    if has_build_arch and has_build_indep:
        sys.exit(0)

    if any([line.lstrip(b' -').startswith(b'include ')
            for line in editor.makefile.dump_lines()]):
        # No handling of includes for the moment.
        sys.exit(0)

    archs = get_archs()
    if not has_build_indep:
        issue = LintianIssue(
            'source', 'debian-rules-missing-recommended-target',
            info='build-indep')
        if issue.should_fix():
            added.append('build-indep')
            editor.makefile.add_rule(
                b'build-indep',
                components=([b'build'] if 'all' in archs else None))
            issue.report_fixed()
    if not has_build_arch:
        added.append('build-arch')
        issue = LintianIssue(
            'source', 'debian-rules-missing-recommended-target',
            info='build-arch')
        if issue.should_fix():
            editor.makefile.add_rule(
                b'build-arch',
                components=([b'build'] if (archs - set(['all'])) else None))
            issue.report_fixed()

    if not added:
        sys.exit(0)

    try:
        phony_rule = list(editor.makefile.iter_rules(b'.PHONY'))[-1]
    except IndexError:
        pass
    else:
        for c in added:
            phony_rule.append_component(c.encode())


if len(added) == 1:
    report_result('Add missing debian/rules target %s.' % added[0])
else:
    report_result('Add missing debian/rules targets %s.' % ', '.join(added))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest


def test_suite():
    names = [
        "changelog",
        "config",
        "debhelper",
        "detect_gbp_dch",
        "lintian",
        "lintian_overrides",
        "multiarch_hints",
        "patches",
        "run",
        "salsa",
        "scrub_obsolete",
        "systemd",
        "vcs",
        "xdg",
        "yaml",
    ]
    module_names = [__name__ + ".test_" + name for name in names]
    module_names.append("upstream_ontologist.tests.test_upstream_ontologist")
    module_names.append("upstream_ontologist.tests.test_vcs")
    module_names.append(__name__ + ".fixers.test_suite")
    loader = unittest.TestLoader()
    return loader.loadTestsFromNames(module_names)



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.xdg."""

from ..xdg import (
    DesktopEntryEditor,
)

from breezy.tests import (
    TestCaseWithTransport,
)


class UpdateDesktopEntryTests(TestCaseWithTransport):
    def test_weird_spacing(self):
        self.build_tree_contents(
            [
                (
                    "foo.desktop",
                    """\
[Desktop Entry]
Name= foo
""",
                )
            ]
        )
        with DesktopEntryEditor("foo.desktop") as updater:
            updater["NewField"] = "foo"
        self.assertFileEqual(
            """\
[Desktop Entry]
Name= foo
NewField=foo
""",
            "foo.desktop",
        )

    def test_modify(self):
        self.build_tree_contents(
            [
                (
                    "foo.desktop",
                    """\
[Desktop Entry]
Name=Foo
""",
                )
            ]
        )
        with DesktopEntryEditor("foo.desktop") as updater:
            updater["NewField"] = "foo"

        self.assertFileEqual(
            """\
[Desktop Entry]
Name=Foo
NewField=foo
""",
            "foo.desktop",
        )



#!/usr/bin/python3

from lintian_brush.fixer import report_result, LintianIssue
from lintian_brush.lintian_overrides import (
    update_overrides,
    LintianOverride,
    load_renamed_tags,
    )


renames = load_renamed_tags()


def rename_override_tags(lineno, override):
    try:
        new_tag = renames[override.tag]
    except KeyError:
        pass  # no rename
    else:
        issue = LintianIssue(
            (override.type, override.package), 'renamed-tag',
            '%s => %s' % (override.tag, new_tag))
        if issue.should_fix():
            issue.report_fixed()
            return LintianOverride(
                package=override.package, archlist=override.archlist,
                type=override.type, tag=new_tag,
                info=override.info)
    return override


update_overrides(rename_override_tags)

report_result("Update renamed lintian tag names in lintian overrides.")



#!/usr/bin/python3

from debian.changelog import Version
import json
import os
from typing import Dict

KEY_PACKAGES = ('debhelper', 'dpkg')

OUTPUT_FILENAME = 'key-package-versions.json'

DEFAULT_UDD_URL = (
    "postgresql://udd-mirror:udd-mirror@udd-mirror.debian.net/udd")

versions: Dict[str, Dict[str, str]]

with open(OUTPUT_FILENAME, 'r') as f:
    versions = json.load(f)


for kp in KEY_PACKAGES:
    versions.setdefault(kp, {})


def update_debian(versions, key_packages):
    import psycopg2
    conn = psycopg2.connect(
        os.environ.get('UDD_URL', DEFAULT_UDD_URL))

    cursor = conn.cursor()
    cursor.execute(
        "SELECT source, release, version from sources where source IN %s",
        (key_packages, ))

    for row in cursor.fetchall():
        versions[row[0]][row[1]] = row[2]


def update_ubuntu(versions, key_packages):
    from launchpadlib.launchpad import Launchpad
    from launchpadlib.uris import LPNET_SERVICE_ROOT
    lp = Launchpad.login_anonymously('lintian-brush', service_root=LPNET_SERVICE_ROOT)
    ubuntu = lp.distributions['ubuntu']
    archive = ubuntu.main_archive
    for series in ubuntu.series:
        print('  .. %s' % series.name)
        for pkg in key_packages:
            if (ubuntu.current_series.name != series.name and
                    series.name in versions[pkg]):
                continue
            ps = archive.getPublishedSources(
                exact_match=True, source_name=pkg, distro_series=series)
            versions[pkg][series.name] = str(max(
                Version(p.source_package_version)
                for p in ps
                if p.pocket == 'Release'))


print('Downloading Debian key package information')
update_debian(versions, KEY_PACKAGES)

print('Downloading Ubuntu key package information')
update_ubuntu(versions, KEY_PACKAGES)


with open(OUTPUT_FILENAME, 'w') as f:
    json.dump(versions, f, indent=4, sort_keys=True)



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    opinionated,
    report_result,
    )


if not opinionated():
    sys.exit(0)

try:
    with open('debian/patches/series', 'r') as f:
        if not f.read().strip():
            os.unlink('debian/patches/series')
except FileNotFoundError:
    pass

report_result('Remove empty debian/patches/series.')



#!/usr/bin/python3

from lintian_brush.fixer import (
    package_is_native,
    report_result,
    fixed_lintian_tag,
    opinionated,
    )
import os
import sys

if not package_is_native():
    # Nothing to do
    sys.exit(0)

if not opinionated():
    sys.exit(0)

try:
    os.unlink('debian/upstream/metadata')
except FileNotFoundError:
    sys.exit(0)
else:
    fixed_lintian_tag('source', 'upstream-metadata-in-native-source', ())

if os.listdir('debian/upstream') == []:
    os.rmdir('debian/upstream')

report_result(
    'Remove debian/upstream/metadata in native source package.',
    certainty='certain')



#!/usr/bin/python3

import os

import re
import sys

from debian.deb822 import PkgRelation
from lintian_brush.fixer import (
    net_access_allowed,
    compat_release,
    report_result,
    LintianIssue,
    )
from debmutate._rules import RulesEditor, discard_pointless_override


compat_release = compat_release()


def previous_release(release):
    import distro_info
    debian = distro_info.DebianDistroInfo()
    if release in (debian.devel(), debian.testing(), 'experimental'):
        return debian.stable()
    releases = debian.get_all()
    try:
        return releases[releases.index(release)-1]
    except ValueError:
        pass
    # TODO(jelmer): Ubuntu?
    return None


VERSION_CMP_SQL = {
    '=': '=',
    '>=': '>=',
    '<=': '<=',
    '>>': '>',
    '<<': '<',
    }


async def package_exists(package, release, version_info):
    if not net_access_allowed():
        try:
            return (
                package in
                os.environ['%s_PACKAGES' % release.upper()].split(','))
        except KeyError:
            return None
    try:
        from lintian_brush.udd import connect_udd_mirror
    except ModuleNotFoundError:
        return None
    async with await connect_udd_mirror() as udd:
        query = 'SELECT True FROM packages WHERE release = $2 AND package = $1'
        args = [package, release]
        if version_info is not None:
            version_cmp, version = version_info
            query += ' AND version %s $3' % VERSION_CMP_SQL[version_cmp]
            args.append(version)
        row = await udd.fetchrow(query, *args)
        return bool(row)


def migration_done(rels):
    import asyncio
    loop = asyncio.get_event_loop()
    previous = previous_release(compat_release)
    if previous is None:
        # We can't determine if the migration is done
        return False
    for rel in rels:
        if len(rel) > 1:
            # Not sure how to handle | Replaces
            return False
        if loop.run_until_complete(package_exists(
                rel[0]['name'], previous, rel[0]['version'])) is not False:
            return False
    return True


def eliminate_dbgsym_migration(line, target):
    if not line.startswith(b'dh_strip'):
        return line

    def rep_dbgsym_migration(m):
        rep = m.group(2).strip(b'"').strip(b"'").decode()
        if '$' in rep:
            # too complicated
            return m.group(0)
        rep = PkgRelation.parse_relations(rep)
        if migration_done(rep):
            issue = LintianIssue(
                'source', 'debug-symbol-migration-possibly-complete',
                '%s (line XX)' % (m.group(0).decode().strip()))
            if issue.should_fix():
                issue.report_fixed()
                return b''
        return m.group(0)

    line = re.sub(
        b'([ \t]+)--dbgsym-migration[= ](\'[^\']+\'|\"[^\"]+\"|[^ ]+)',
        rep_dbgsym_migration, line)

    if line == b'dh_strip || dh_strip':
        line = b'dh_strip'

    return line


if not os.path.exists('debian/rules'):
    sys.exit(2)
with RulesEditor() as editor:
    for rule in editor.makefile.iter_all_rules():
        newlines = []
        for line in list(rule.lines[1:]):
            if line.startswith(b'\t'):
                ret = eliminate_dbgsym_migration(line[1:], rule.target)
                newlines.append(b'\t' + ret)
            else:
                newlines.append(line)
        oldlines = rule.lines
        if oldlines != newlines:
            rule.lines = [rule.lines[0]] + newlines
            discard_pointless_override(editor.makefile, rule, ignore_comments=True)


report_result('Drop transition for old debug package migration.')
